# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#
# pylint:disable=protected-access,too-many-lines,unused-variable

from decimal import Decimal
import datetime
import logging
import os
import warnings
from typing import Any, cast, List, TypeVar

import pytz
from django.conf import settings
from django.db import connections
from django.db.models import Q, Avg, Count, Sum, Min, Max, Model, query as models_query, functions
from django.test import TestCase, override_settings
from django.utils import timezone
from django.utils.crypto import get_random_string

import salesforce
from salesforce import router
from salesforce.backend import DJANGO_50_PLUS
from salesforce.backend.test_helpers import (  # noqa pylint:disable=unused-import
    expectedFailure, expectedFailureIf, skip, skipUnless, strtobool)
from salesforce.backend.test_helpers import (
    current_user, default_is_sf, sf_alias, uid_version as uid,
    QuietSalesforceErrors, LazyTestMixin)
from salesforce.dbapi.driver import ApiUsage
from salesforce.dbapi.exceptions import SalesforceWarning
from salesforce.dbapi.test_helpers import PatchedSfConnection
from salesforce.models import SalesforceModel
from salesforce.testrunner.example.models import (
        Attachment, Account, Contact, Lead, User,
        ApexEmailNotification, BusinessHours, Campaign, CronTrigger,
        Opportunity, OpportunityContactRole,
        Product, Pricebook, PricebookEntry, Note, Task,
        Organization, models_template, Test,
        )

log = logging.getLogger(__name__)

_M = TypeVar('_M', bound=Model)

QUIET_KNOWN_BUGS = strtobool(os.getenv('QUIET_KNOWN_BUGS', 'false'))
test_email = 'test-djsf-unittests%s@example.com' % uid
sf_databases = [db for db in connections if router.is_sf_database(db)]

_sf_tables = []  # type: List[str]


def sf_tables() -> List[str]:
    if not _sf_tables and default_is_sf:
        for x in connections[sf_alias].introspection.table_list_cache['sobjects']:
            _sf_tables.append(x['name'])
    return _sf_tables


# The field name 'id' or 'Id' in these universal tests that should work with
# both variants of SF_PK. The name 'pk' is prefered here.
# The name 'Id' is used in SQL code, but the name 'id' would work just as well.

def refresh(obj: _M) -> _M:
    """Get the same object refreshed from the same db.
    """
    db = obj._state.db
    qs: models_query.QuerySet[_M] = type(obj).objects.using(db)
    return qs.get(pk=obj.pk)


class BasicSOQLRoTest(TestCase, LazyTestMixin):
    """Tests that need no setUp/tearDown"""
    # pylint:disable=no-self-use,pointless-statement

    databases = '__all__'

    @classmethod
    def setUpClass(cls) -> None:
        """Add contact if less than 2 exist"""
        super(BasicSOQLRoTest, cls).setUpClass()
        if User.objects.count() == 0:
            User.objects.create(Username=current_user)                       # pragma: no test
        some_accounts = list(Account.objects.all()[:1])
        if not some_accounts:
            some_accounts = [Account.objects.create(Name='sf_test account_0')]
        some_contacts = Contact.objects.exclude(account__isnull=True).filter(name__gt='A')[:2]
        if len(some_contacts) < 2:
            for i in range(2 - len(some_contacts)):
                Contact.objects.create(first_name='sf_test demo', last_name='Test %d' % i,
                                       account=some_accounts[0])

    def test_raw(self) -> None:
        """Read two contacts by raw.

        (At least 3 manually created Contacts must exist before these read-only tests.)
        """
        with self.lazy_assert_n_requests(1):
            contacts = list(Contact.objects.raw(
                "SELECT Id, LastName, FirstName FROM Contact "
                "LIMIT 2"))
        with self.lazy_assert_n_requests(0):
            self.assertEqual(len(contacts), 2)
            # It had a side effect that the same assert failed second times.
            self.assertEqual(len(contacts), 2)
            '%s' % contacts[0].__dict__  # Check that all fields are accessible
        self.lazy_check()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_raw_translations(self) -> None:
        """Read a Contact raw and translate it to Lead fields."""
        contact = Contact.objects.all()[0]
        # `translation` dictionary key is lowercase
        false_lead_raw = list(Lead.objects.raw(
            "SELECT Id, LastName FROM Contact WHERE Id=%s", params=[contact.pk],
            translations={'lastname': 'Company'}))
        self.assertEqual(len(false_lead_raw), 1)
        self.assertEqual(false_lead_raw[0].Company, contact.last_name)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_raw_foreignkey_id(self) -> None:
        """Get the first two contacts by raw query with a ForeignKey id field.
        """
        with self.lazy_assert_n_requests(1):
            contacts = list(Contact.objects.raw(
                "SELECT Id, LastName, FirstName, OwnerId FROM Contact "
                "LIMIT 2"))
        len(contacts)
        self.assertEqual(len(contacts), 2)
        with self.lazy_assert_n_requests(0):
            '%s' % contacts[0].__dict__  # Check that all fields are accessible
        with self.lazy_assert_n_requests(1):
            self.assertIn('@', contacts[0].owner.Email)
        self.lazy_check()

    def test_select_all(self) -> None:
        """Get the first two contact records.
        """
        contacts = Contact.objects.all()[0:2]
        self.assertEqual(len(contacts), 2)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_foreign_key(self) -> None:
        """Verify that the owner of an Contact is the currently logged admin.
        """
        current_sf_user = User.objects.get(Username=current_user)
        test_contact = Contact(first_name='sf_test', last_name='my')
        test_contact.save()
        try:
            contact = Contact.objects.filter(owner=current_sf_user)[0]
            user = contact.owner
            # This user can be e.g. 'admins@example.com.prod001'.
            self.assertEqual(user.Username, current_user)
        finally:
            test_contact.delete()

    def test_foreign_key_column(self) -> None:
        """Verify filtering by a column of related parent object.
        """
        test_account = Account(Name='sf_test account')
        test_account.save()
        test_contact = Contact(first_name='sf_test', last_name='my', account=test_account)
        test_contact.save()
        try:
            contacts = Contact.objects.filter(account__Name='sf_test account')
            self.assertEqual(len(contacts), 1)
        finally:
            test_contact.delete()
            test_account.delete()

    def test_select_related(self) -> None:
        """Verify that select_related does not require additional queries.
        """
        test_account = Account(Name='sf_test account')
        test_account.save()
        test_contact = Contact(first_name='sf_test', last_name='my', account=test_account)
        test_contact.save()
        try:
            with self.lazy_assert_n_requests(1):
                qs = Contact.objects.filter(account__Name='sf_test account').select_related('account')
                contacts = list(qs)
            with self.lazy_assert_n_requests(0):
                [x.account.Name for x in contacts]
            self.assertGreaterEqual(len(contacts), 1)
            self.lazy_check()
        finally:
            test_contact.delete()
            test_account.delete()

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_select_related_child_subquery(self) -> None:
        """Test select_related with a subquery by children objects.
        """
        with self.lazy_assert_n_requests(1):
            qs = Account.objects.filter(contact__isnull=False, Owner__isnull=False).select_related('Owner')
            if 'AND Contact.Id' in str(qs.query):
                sql_end = 'FROM Contact WHERE (Contact.Account.OwnerId != null AND Contact.Id != null)'
            else:
                sql_end = 'FROM Contact WHERE (Contact.Id != null AND Contact.Account.OwnerId != null)'
            self.assertIn(sql_end, str(qs.query))
            self.assertGreater(len(list(qs)), 0)
            self.assertGreater(qs[0].Owner.Username, '')

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_select_related_child_filter(self) -> None:
        """Test select_related with a subquery by children objects.
        """
        with self.lazy_assert_n_requests(1):
            subquery = Contact.objects.filter().values('account_id')
            qs = Account.objects.filter(pk__in=subquery).select_related('Owner')
            self.assertGreater(len(qs), 0)
            self.assertGreater(qs[0].Owner.Username, '')
        soql = str(qs.query)
        self.assertTrue(soql.endswith("FROM Account WHERE Account.Id IN (SELECT Contact.AccountId FROM Contact)"))

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_select_related_child_exclude(self) -> None:
        """Test select_related with a subquery by children objects.
        """
        # We use 'not_in' lookup, because this is not supported: 'exclude(pk__in=subsuery)'
        #     Account.objects.exclude(contact__isnull=True)

        with self.lazy_assert_n_requests(1):
            subquery = Contact.objects.values('account_id')
            qs = Account.objects.filter(pk__not_in=subquery).select_related('Owner')
            list(qs)
        soql = str(qs.query)
        self.assertTrue(soql.endswith("FROM Account WHERE Account.Id NOT IN (SELECT Contact.AccountId FROM Contact)"))

        # the same without 'not_in' lookup by two requests and exclude()
        with self.lazy_assert_n_requests(2):
            sub_ids = Contact.objects.exclude(account=None).values_list('account_id', flat=True)[:100]
            qs = Account.objects.exclude(pk__in=list(sub_ids)).select_related('Owner')
            list(qs)
        soql = str(qs.query)
        self.assertRegex(soql, r"FROM Account WHERE \(NOT \(Account.Id IN \((?:NULL, )?'001\w{15}'")

    def test_not_eq(self) -> None:
        qs = Contact.objects.filter(email__not_eq='')
        self.assertTrue(all(x.email > '' for x in qs))

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_one_to_one_field(self) -> None:
        # test that a OneToOneField is a unique field
        self.assertEqual(ApexEmailNotification._meta.get_field('user').unique, True)  # type: ignore[attr-defined]

        current_sf_user = User.objects.get(Username=current_user)
        orig_objects = list(ApexEmailNotification.objects.filter(
            Q(user=current_sf_user) | Q(email='apex.bugs@example.com')))
        _ = orig_objects  # noqa
        new_u, new_e = None, None
        try:
            notifier_u = current_sf_user.apex_email_notification
        except ApexEmailNotification.DoesNotExist:
            notifier_u = new_u = ApexEmailNotification(user=current_sf_user)
            notifier_u.save()
        try:
            notifier_e = ApexEmailNotification.objects.get(email='apex.bugs@example.com')
        except ApexEmailNotification.DoesNotExist:
            notifier_e = new_e = ApexEmailNotification(email='apex.bugs@example.com')
            notifier_e.save()

        try:
            # test 1b is unique value
            duplicate = ApexEmailNotification(user=current_sf_user)
            # the method self.assertRaise was too verbose about exception
            try:
                with QuietSalesforceErrors(sf_alias):
                    duplicate.save()
            except salesforce.backend.base.SalesforceError as exc:
                # TODO uncovered line, maybe bug
                self.assertEqual(exc.data[0]['errorCode'], 'DUPLICATE_VALUE')  # type: ignore
            else:
                self.assertRaises(salesforce.backend.base.SalesforceError, duplicate.save)

            # test 2: the reverse relation is a value, not a set
            users = User.objects.exclude(apex_email_notification__user=None)
            self.assertIn(current_user, [x.Username for x in users])

            # test 3: relation to the parent
            result = ApexEmailNotification.objects.filter(user__Username=notifier_u.user.Username)
            self.assertEqual(len(result), 1)
            # sometimes mypy error: '"ApexEmailNotification" has no attribute "user_id"'
            # probably caused by .mypy_cache  # TODO report it if reproduced with a new django-stubs version
            self.assertEqual(result[0].user_id, notifier_u.user_id)
        finally:
            if new_u:
                new_u.delete()
            if new_e:
                new_e.delete()
        list(User.objects.exclude(apex_email_notification=None))

    def test_update_date(self) -> None:
        """Test updating a date.
        """
        # removed microseconds only for easy compare in the test, no problem
        now = timezone.now().replace(microsecond=0)
        contact = Contact(first_name='sf_test', last_name='my')
        contact.save()
        contact = refresh(contact)
        try:
            contact.email_bounced_date = now
            contact.save()
            if not settings.USE_TZ:
                now = timezone.make_aware(now, timezone.get_current_timezone())
            self.assertEqual(refresh(contact).email_bounced_date, now)
        finally:
            contact.delete()

    def test_insert_date(self) -> None:
        """Test inserting a date.
        """
        now = timezone.now().replace(microsecond=0)
        contact = Contact(first_name='Joe', last_name='Freelancer', email_bounced_date=now)
        contact.save()
        try:
            if not settings.USE_TZ:
                now = timezone.make_aware(now, timezone.get_current_timezone())
            self.assertEqual(refresh(contact).email_bounced_date, now)
        finally:
            contact.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    @skipUnless(DJANGO_50_PLUS, "Django >= 5.0 is required for the test.")
    def test_default_specified_by_sf(self) -> None:
        """Verify insert of object with a field with default value on create by SF.

        (The default is used only for a field unspecified in SF REST API, but
        not for None or any similar value. It was a pain for some unimportant
        foreign keys that don't accept null.)
        """
        # Verify a smart default is used.
        contact = Contact(first_name='sf_test', last_name='my')
        contact.save()
        try:
            self.assertEqual(refresh(contact).owner.Username, current_user)
        finally:
            contact.delete()
        # Verify that an explicit value is possible for this field.
        other_user_obj = (User.objects.exclude(Username=current_user)
                          .filter(IsActive=True, UserType='Standard')[0])
        contact = Contact(first_name='sf_test', last_name='your',
                          owner=other_user_obj)
        contact.save()
        try:
            self.assertEqual(refresh(contact).owner.Username, other_user_obj.Username)
        finally:
            contact.delete()

    def test_not_null_related(self) -> None:
        """Verify conditions `isnull` for foreign keys: filter(Account=None)

        filter(Account__isnull=True) and nested in Q(...) | Q(...).
        """
        test_contact = Contact(first_name='sf_test', last_name='my')
        test_contact.save()
        try:
            contacts = Contact.objects.filter(
                Q(account__isnull=True) | Q(account=None),
                account=None,
                account__isnull=True,
                first_name='sf_test'
            )
            self.assertEqual(len(contacts), 1)
        finally:
            test_contact.delete()

    def test_unicode(self) -> None:
        """Make sure weird unicode breaks properly.
        """
        test_lead = Lead(FirstName='\u2603', LastName="Unittest Unicode",
                         Email='test-djsf-unicode-email@example.com',
                         Company="Some company")
        test_lead.save()
        try:
            self.assertEqual(refresh(test_lead).FirstName, '\u2603')
        finally:
            test_lead.delete()

    def test_date_comparison(self) -> None:
        """Test that date comparisons work properly.
        """
        today = datetime.datetime(2013, 8, 27)
        if settings.USE_TZ:
            today = timezone.make_aware(today, pytz.utc)
        yesterday = today - datetime.timedelta(days=1)
        tomorrow = today + datetime.timedelta(days=1)
        contact = Contact(first_name='sf_test' + uid, last_name='date',
                          email_bounced_date=today)
        contact.save()
        try:
            contacts1 = Contact.objects.filter(email_bounced_date__gt=yesterday, first_name='sf_test' + uid)
            self.assertEqual(len(contacts1), 1)
            contacts2 = Contact.objects.filter(email_bounced_date__gt=tomorrow, first_name='sf_test' + uid)
            self.assertEqual(len(contacts2), 0)
        finally:
            contact.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_insert(self) -> None:
        """Create a lead record, and make sure it ends up with a valid Salesforce ID.
        """
        test_lead = Lead(FirstName="User", LastName="Unittest Inserts",
                         Email='test-djsf-inserts-email@example.com',
                         Company="Some company")
        test_lead.save()
        try:
            self.assertEqual(len(test_lead.pk), 18)
        finally:
            test_lead.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_double_save_default(self) -> None:
        """Double save without refresh of a field with default=DEFAULTED_ON_CREATE

        This should write a warning but not cause a problem.
        """
        oppo = Opportunity(name='test op', stage='Prospecting', close_date=datetime.date.today())
        try:
            oppo.save()
            with self.assertWarns(SalesforceWarning) as cm:
                # should save a DEFAULTED_ON_CREATE field 'probability' on update, but with warning
                oppo.save()
                message = cm.warnings[0].message
                self.assertIn("saved again with DEFAULTED_ON_CREATE value", message.args[0])

            oppo.save(update_fields=['name', 'stage'])
            with self.assertWarns(SalesforceWarning):
                oppo.save(update_fields=['name', 'stage', 'probability'])

            # a normal value can be saved on update
            oppo.probability = '25'  # percent
            oppo.save()
        finally:
            oppo.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    @skipUnless(DJANGO_50_PLUS, "Django >= 5.0 is required for db_default fields")
    def test_double_save_db_default(self) -> None:
        """Double save without refresh of a field with db_default=DEFAULTED_ON_CREATE
        """
        contact = Contact(last_name='sf_test')
        try:
            contact.save()
            contact.save()
        finally:
            contact.delete()

    def test_delete(self) -> None:
        """Create a lead record, then delete it, and make sure it's gone.
        """
        test_lead = Lead(FirstName="User", LastName="Unittest Deletes",
                         Email='test-djsf-delete-email@example.com',
                         Company="Some company")
        test_lead.save()
        test_lead.delete()

        self.assertRaises(Lead.DoesNotExist, Lead.objects.get, Email='test-djsf-delete-email@example.com')

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_decimal_precision(self) -> None:
        """Verify the exact precision of the saved and retrived DecimalField
        """
        product = Product(Name="test Product")
        product.save()

        # The price for a product must be set in the standard price book.
        # http://www.salesforce.com/us/developer/docs/api/Content/sforce_api_objects_pricebookentry.htm
        pricebook = Pricebook.objects.get(Name="Standard Price Book")
        unit_price = Decimal('1234.56')
        saved_pricebook_entry = PricebookEntry(Product2=product, Pricebook2=pricebook,
                                               UnitPrice=unit_price, UseStandardPrice=False)
        saved_pricebook_entry.save()
        retrieved_pricebook_entry = PricebookEntry.objects.get(pk=saved_pricebook_entry.pk)

        try:
            self.assertEqual(saved_pricebook_entry.UnitPrice, unit_price)
            self.assertEqual(retrieved_pricebook_entry.UnitPrice, unit_price)
        finally:
            retrieved_pricebook_entry.delete()
            product.delete()

    def test_zero_decimal_places(self) -> None:
        """Test that DecimalField with decimal_places=0 is correctly parsed"""
        campaign = Campaign(name='test something', number_sent=3)
        campaign.save()
        try:
            ret = Campaign.objects.filter(number_sent=3)
            # should be parsed without ".0"
            val = ret[0].number_sent
            self.assertEqual(repr(val), "Decimal('3')")
            self.assertEqual(str(val), "3")
        finally:
            campaign.delete()

    def test_simple_custom_object(self) -> None:
        """Create, read and delete a simple custom object `django_Test__c`.
        """
        if 'django_Test__c' not in sf_tables():
            self.skipTest("Not found custom object 'django_Test__c'")
        results = Test.objects.all()[0:1]
        obj = Test(test_text='sf_test')
        obj.save()
        attachment = Attachment(name='test attachment', parent=obj)
        attachment.save()
        try:
            results = Test.objects.all()[0:1]
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0].test_text, 'sf_test')
        finally:
            attachment.delete()
            obj.delete()

    def test_datetime_miliseconds(self) -> None:
        """Verify that a field with milisecond resolution is readable.
        """
        triggers = CronTrigger.objects.all()
        if not triggers:
            self.skipTest("No object with milisecond resolution found.")
        self.assertTrue(isinstance(triggers[0].PreviousFireTime, datetime.datetime))
        if all(trigger.PreviousFireTime.microsecond == 0 for trigger in triggers):
            self.skipTest("Miliseconds == 0; That is possible with probability 0.1%, not error")

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_time_field(self) -> None:
        """Test a TimeField (read, modify, verify).
        """
        obj_orig = BusinessHours.objects.all()[0]
        obj = refresh(obj_orig)
        self.assertTrue(isinstance(obj.MondayStartTime, datetime.time))
        obj.MondayStartTime = datetime.time(23, 59)
        obj.save()
        obj = refresh(obj)
        try:
            self.assertEqual(obj.MondayStartTime, datetime.time(23, 59))
        finally:
            obj_orig.save()

    def test_account_insert_delete(self) -> None:
        """Test insert and delete an account (normal or personal SF config)
        """
        if getattr(settings, 'SF_EXAMPLE_PERSON_ACCOUNT_ACTIVATED', False):
            test_account = Account(FirstName='IntegrationTest',  # type: ignore[misc] # skip this branch
                                   LastName='Account')
        else:
            test_account = Account(Name='IntegrationTest Account')
        test_account.save()
        try:
            accounts = Account.objects.filter(Name='IntegrationTest Account')
            self.assertEqual(len(accounts), 1)
        finally:
            test_account.delete()

    def test_similarity_filter_operators(self) -> None:
        """Test filter operators that use LIKE 'something%' and similar.
        """
        User.objects.get(Username__exact=current_user)
        User.objects.get(Username__iexact=current_user.upper())
        User.objects.get(Username__contains=current_user[1:-1])
        User.objects.get(Username__icontains=current_user[1:-1].upper())
        User.objects.get(Username__startswith=current_user[:-1])
        User.objects.get(Username__istartswith=current_user[:-1].upper())
        User.objects.get(Username__endswith=current_user[1:])
        User.objects.get(Username__iendswith=current_user[1:].upper())
        # Operators regex and iregex not tested because they are not supported.

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_bulk_create(self) -> None:
        """Create two Contacts by one request in one command, find them.
        """
        account = Account(Name='test bulk')
        account.save()
        try:
            objects = [Contact(last_name='sf_test a', account=account),
                       Contact(last_name='sf_test b', account=account)]

            with self.lazy_assert_n_requests(1):
                ret = Contact.objects.bulk_create(objects)
                # test that can return ids from bulk_create
                self.assertEqual(len(set(x.pk for x in ret if x.pk)), 2)

            self.assertEqual(len(Contact.objects.filter(account=account)), 2)
            self.lazy_check()
        finally:
            account.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_queryset_update(self) -> None:
        """Update two Accounts by one request, after searching them by one request.
        """
        account_0, account_1 = [Account(Name='test' + uid), Account(Name='test' + uid)]
        account_0.save()
        account_1.save()
        try:
            # 2x update with different filters by primary key: 1 request per update
            with self.lazy_assert_n_requests(1):
                Account.objects.filter(pk=account_0.pk).update(Name="test2" + uid)
            with self.lazy_assert_n_requests(1):
                Account.objects.filter(pk__in=[account_1.pk]).update(Name="test2" + uid)

            # 1x update of 2 records + 1 more complicated filter == 2 requests
            with self.lazy_assert_n_requests(2):
                Account.objects.filter(pk__in=Account.objects.filter(Name='test2' + uid)).update(Name="test3" + uid)
            self.assertEqual(Account.objects.filter(Name='test3' + uid).count(), 2)
            self.lazy_check()
        finally:
            account_0.delete()
            account_1.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_queryset_update_foreign_key(self) -> None:
        contact = Contact.objects.exclude(account=None)[0]
        Contact.objects.filter(pk=contact.pk).update(account=contact.account)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_queryset_none_filter_update(self) -> None:
        """Test that this can be compiled and the SOQL is valid"""
        Contact.objects.filter(pk=None).update(last_name='a')

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_bulk_delete_and_update(self) -> None:
        """Delete two Accounts by one request.
        """
        account_0, account_1 = [Account(Name='test' + uid), Account(Name='test' + uid)]
        account_0.save()
        account_1.save()
        try:
            result_count = 2
            with self.lazy_assert_n_requests(2):
                ret: Any = Account.objects.filter(Name='test' + uid).delete()
            self.assertEqual(ret, (2, {'example.Account': 2}))
            result_count = Account.objects.filter(Name='test' + uid).count()
            self.assertEqual(result_count, 0)
            self.lazy_check()
        finally:
            if result_count:
                account_0.delete()
                account_1.delete()
        try:
            nn = 200
            pks = []
            objects = [Lead(Company='sf_test lead', LastName='name_{}'.format(i))
                       for i in range(nn)]
            with self.lazy_assert_n_requests(1):
                ret = Lead.objects.bulk_create(objects)
            pks = [x.pk for x in objects if x.pk]
            with self.lazy_assert_n_requests(1):
                ret = Lead.objects.filter(pk__in=pks).update(Company='sf_test lead_2')
        finally:
            if pks:
                with self.lazy_assert_n_requests(1):
                    ret = Lead.objects.filter(pk__in=pks).delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_bulk_delete_all(self) -> None:
        PricebookEntry.objects.filter(pk__gt='').delete()

    def test_escape_single_quote(self) -> None:
        """Test single quotes in strings used in a filter

        Verity that they are escaped properly.
        """
        account_name = '''Dr. Evil's Giant\\' "Laser", LLC'''
        account = Account(Name=account_name)
        account.save()
        try:
            self.assertTrue(Account.objects.filter(Name=account_name).exists())
        finally:
            account.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_escape_single_quote_in_raw_query(self) -> None:
        """Test that manual escaping within a raw query is not double escaped.
        """
        account_name = '''test Dr. Evil's Giant\\' "Laser", LLC'''
        account = Account(Name=account_name)
        account.save()

        manually_escaped = '''test Dr. Evil\\'s Giant\\\\\\' "Laser", LLC'''
        try:
            retrieved_account = Account.objects.raw(
                "SELECT Id, Name FROM Account WHERE Name = '%s'" % manually_escaped)[0]
            self.assertEqual(account_name, retrieved_account.Name)
        finally:
            account.delete()

    def test_raw_query_empty(self) -> None:
        """Test that the raw query works even for queries with empty results.

        Raw queries with supported empty results are an improvement over normal
        Django. It is useful if some unimplemented features
        of django-salesforce are solved by writing a raw query.
        """
        len(list(Contact.objects.raw("SELECT Id, FirstName FROM Contact WHERE FirstName='nonsense'")))

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_range_simple(self) -> None:
        """Test simple range filters".
        """
        qs = Contact.objects.filter(Q(name__range=('c', 'e')))
        soql = qs.query.get_compiler('salesforce').as_sql()[0]
        self.assertIn("(Contact.Name >= %s AND Contact.Name <= %s)", soql)
        len(qs)

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_range_combined(self) -> None:
        """Test combined filters "a OR b AND c".
        """
        qs = Contact.objects.filter(Q(name='a') | Q(name__range=('c', 'e')))
        soql = qs.query.get_compiler('salesforce').as_sql()[0]
        self.assertIn("Contact.Name = %s OR (Contact.Name >= %s AND Contact.Name <= %s)", soql)
        len(qs)

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_range_lookup(self) -> None:
        """Get the test opportunity record by range condition.
        """
        test_opportunity = Opportunity(
                        name="Example Opportunity",
                        close_date=datetime.date(year=2015, month=7, day=30),
                        stage="Prospecting",
                        amount=130000.00
        )
        test_opportunity.save()
        try:
            # Test date objects
            start_date = datetime.date(year=2015, month=7, day=29)
            end_date = datetime.date(year=2015, month=8, day=1)
            oppy = Opportunity.objects.filter(close_date__range=(start_date, end_date))[0]
            self.assertEqual(oppy.name, 'Example Opportunity')
            self.assertEqual(oppy.stage, 'Prospecting')

            # Test datetime objects (now +- 10 minutes for clock inaccuracy)
            start_time = timezone.now() - datetime.timedelta(seconds=600)
            end_time = timezone.now() + datetime.timedelta(seconds=600)
            opportunities = Opportunity.objects.filter(created_date__range=(start_time, end_time))[:1]
            self.assertEqual(len(opportunities), 1, "Failed range filter or maybe incorrectly set clock")
            oppy = opportunities[0]
            self.assertEqual(oppy.name, 'Example Opportunity')
            self.assertEqual(oppy.stage, 'Prospecting')

            # Test DecimalField
            low_amount, high_amount = 100000.00, 140000.00
            oppy = Opportunity.objects.filter(amount__range=(low_amount, high_amount))[0]
            self.assertEqual(oppy.amount, 130000.00)

            # Test CharField
            low_letter, high_letter = 'E', 'G'
            oppy = Opportunity.objects.filter(name__range=(low_letter, high_letter))[0]
            self.assertEqual(oppy.name, 'Example Opportunity')
        finally:
            test_opportunity.delete()

    def test_combined_international(self) -> None:
        """Test combined filters with international characters.
        """
        # This is OK for long time
        len(Contact.objects.filter(Q(first_name='\xe1') & Q(last_name='\xe9')))
        # This was recently fixed
        len(Contact.objects.filter(Q(first_name='\xe1') | Q(last_name='\xe9')))
        len(Contact.objects.filter(Q(first_name='\xc3\xa1') | Q(last_name='\xc3\xa9')))

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_aggregate_query(self) -> None:
        """Test for different aggregate function.
        """
        test_product = Product(Name='test soap')
        test_product.save()
        test_product2 = Product(Name='test brush')
        test_product2.save()
        pricebook = Pricebook.objects.get(Name="Standard Price Book")
        PricebookEntry(Product2=test_product, Pricebook2=pricebook,
                       UseStandardPrice=False, UnitPrice=Decimal(100)).save()
        PricebookEntry(Product2=test_product2, Pricebook2=pricebook,
                       UseStandardPrice=False, UnitPrice=Decimal(80)).save()
        try:
            x_products = PricebookEntry.objects.filter(Name__startswith='test ')
            result = x_products.aggregate(Sum('UnitPrice'), Count('UnitPrice'), Avg('UnitPrice'),
                                          Min('UnitPrice'), Max('UnitPrice'))
            self.assertDictEqual(result, {'UnitPrice__sum': 180, 'UnitPrice__count': 2, 'UnitPrice__avg': 90.0,
                                          'UnitPrice__min': 80, 'UnitPrice__max': 100})
        finally:
            # dependent PricebookEntries are just deleted automatically by SF
            test_product.delete()
            test_product2.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_z_big_query(self) -> None:
        """Test a big query that will be splitted to more requests.

        Test it as late as possible.
        """
        cur = connections[sf_alias].cursor()
        cursor = cur.cursor
        cursor.execute("SELECT Id, Name FROM Lead", query_all=True)
        first_lead = cursor.fetchone()
        if first_lead:
            first_chunk_len = len(cursor._chunk)
            leads_list = [first_lead] + cursor.fetchall()
        else:
            first_chunk_len = 0
            leads_list = []
        if first_chunk_len == len(leads_list):
            self.assertLessEqual(len(leads_list), 2000, "Obsoleted constants")
            log.info("Not enough Leads accumulated (currently %d including deleted) "
                     "in the last two weeks that are necessary for splitting the "
                     "query into more requests. Number 2001 is enough.",
                     len(leads_list))
            self.skipTest("Not enough Leads found for big query test")

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_cursor_execute_fetch(self) -> None:
        """Get results by cursor.execute(...); fetchone(), fetchmany(), fetchall()
        """
        sql = "SELECT Id, LastName, FirstName, OwnerId FROM Contact LIMIT 2"
        cursor = connections[sf_alias].cursor()
        cursor.execute(sql)
        contacts = cursor.fetchall()
        self.assertEqual(len(contacts), 2)
        self.assertTrue(contacts[0][3].startswith('005'), "OwnerId must be an User.")
        cursor.execute(sql)
        self.assertEqual(cursor.fetchone(), contacts[0])
        self.assertEqual(cursor.fetchmany(), contacts[1:])

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_cursor_execute_aggregate(self) -> None:
        """Verify that aggregate queries can be executed directly by SOQL.
        """
        # Field 'Id' is very useful for COUNT over non empty fields.
        sql = "SELECT LastName, COUNT(Id) FROM Contact GROUP BY LastName LIMIT 2"
        cursor = connections[sf_alias].cursor()
        cursor.execute(sql)
        contact_aggregate = cursor.fetchone()
        self.assertEqual([x[0] for x in cursor.description], ['LastName', 'expr0'])
        self.assertEqual([type(x) for x in contact_aggregate], [str, int])
        self.assertGreaterEqual(contact_aggregate[1], 1)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_cursor_list_plain_count(self) -> None:
        """Test that a plain COUNT() works with a list cursor.
        """
        cur = connections[sf_alias].cursor()
        cursor = cur.cursor
        cursor.execute("SELECT COUNT() FROM Contact")
        ret = cursor.fetchall()
        count = ret[0][0]
        self.assertTrue(isinstance(count, int))
        self.assertEqual(ret,  [(count,)])

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_cursor_dict_plain_count(self) -> None:
        """Test that a plain COUNT() works with a dict cursor.
        """
        cur = connections[sf_alias]._cursor(name='dict')
        cursor = cur.cursor
        cursor.execute("SELECT COUNT() FROM Contact")
        ret = cursor.fetchall()
        self.assertEqual(ret,  [{'count': ret[0]['count']}])

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_group_by_compile(self) -> None:
        """Test that group annotations can be correctly compiled and executed"""
        qs = (Contact.objects.filter(account__Name__gt='').order_by()
              .values('account_id').annotate(cnt=Count('pk'))
              )
        soql, params = qs.query.get_compiler('salesforce').as_sql()
        expected_soql = (
            'SELECT Contact.AccountId, COUNT(Contact.Id) cnt FROM Contact '
            'WHERE Contact.Account.Name > %s GROUP BY Contact.AccountId'
        )
        self.assertEqual(soql, expected_soql)
        self.assertEqual(params, ('',))
        self.assertEqual(set(list(qs)[0].keys()), {'account_id', 'cnt'})
        list(qs)

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_having_compile(self) -> None:
        """Test GROUP BY ... HAVING compile for a ManyToMany field"""
        qs = (Contact.objects.values('testdetail__parent__test_bool').order_by()
              .annotate(val=Max('testdetail__parent__test_text'))
              .filter(val__contains='b', val__gt='a')
              )
        expected_soql = (
            "SELECT django_Test_detail__c.Parent__r.TestBool__c,"
            " MAX(django_Test_detail__c.Parent__r.TestText__c) val "
            "FROM django_Test_detail__c "
            "GROUP BY django_Test_detail__c.Parent__r.TestBool__c "
            "HAVING (MAX(django_Test_detail__c.Parent__r.TestText__c) LIKE '%b%'"
            " AND MAX(django_Test_detail__c.Parent__r.TestText__c) > 'a')"
        )
        self.assertEqual(str(qs.query), expected_soql)
        if not ('django_Test__c' in sf_tables() and 'django_Test_detail__c' in sf_tables()):
            self.skipTest("Not found custom object 'django_Test__c'")
        list(qs)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_order_by_compile(self) -> None:
        """Test order_by() compile for a field by ManyToMany relationship"""
        qs = Contact.objects.values('pk').order_by('testdetail__parent__test_text')
        soql, params = qs.query.get_compiler('salesforce').as_sql()
        expected_soql = (
            'SELECT django_Test_detail__c.Contact__r.Id FROM django_Test_detail__c '
            'ORDER BY django_Test_detail__c.Parent__r.TestText__c ASC'
        )
        self.assertEqual(soql, expected_soql)
        self.assertEqual(params, ())
        if not ('django_Test__c' in sf_tables() and 'django_Test_detail__c' in sf_tables()):
            self.skipTest("Not found custom object 'django_Test__c'")
        list(qs)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_recursive_parent(self) -> None:
        qs = Campaign.objects.values('name', 'parent__name', 'parent__parent__name').annotate()
        expected = "SELECT Campaign.Name, Campaign.Parent.Name, Campaign.Parent.Parent.Name FROM Campaign"
        self.assertEqual(str(qs.query), expected)
        list(qs)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_count_distinct(self) -> None:
        """Test Count(some_field, distinct=True)"""
        qs = (Contact.objects.order_by().values('account_id')
              .annotate(cnt=Count('first_name', distinct=True))
              )
        soql, params = qs.query.get_compiler('salesforce').as_sql()
        expected_soql = (
            'SELECT Contact.AccountId, COUNT_DISTINCT(Contact.FirstName) cnt FROM Contact '
            'GROUP BY Contact.AccountId'
        )
        self.assertEqual(soql, expected_soql)
        self.assertEqual(params, ())
        self.assertEqual(set(list(qs)[0].keys()), {'account_id', 'cnt'})
        list(qs)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_count_asterisk(self) -> None:
        """Test Count('*')"""
        with PatchedSfConnection(sf_alias, verbs=['use_debug_info']):
            ret = Contact.objects.aggregate(cnt=Count('*'))
            debug_info = connections[sf_alias].connection.debug_info
        self.assertEqual(debug_info['soql'][:2], ('SELECT COUNT(Id) cnt FROM Contact', []))
        self.assertTrue('cnt' in ret)

        qs = Contact.objects.order_by().values('account_id').annotate(cnt=Count('*'))
        soql, params = qs.query.get_compiler('salesforce').as_sql()
        self.assertEqual(soql, 'SELECT Contact.AccountId, COUNT(Id) cnt FROM Contact GROUP BY Contact.AccountId')
        list(qs)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_errors(self) -> None:
        """Test for improving code coverage.
        """
        # broken query raises exception
        bad_queryset = Lead.objects.raw("select XYZ from Lead")
        with QuietSalesforceErrors(sf_alias):
            self.assertRaises(salesforce.backend.base.SalesforceError, list, bad_queryset)

    def test_queryset_values(self) -> None:
        """Test list of dict qs.values()
        """
        values = Contact.objects.values()[:2]
        self.assertEqual(len(values), 2)
        self.assertIn('first_name', values[0])
        self.assertGreater(len(values[0]), 3)

    def test_queryset_values_names(self) -> None:
        """Test list of dict qs.values(*names) and list of tuples qs.values_list()
        """
        tmp = Contact.objects.values('pk', 'last_name')
        tmp[0]
        values = Contact.objects.values('pk', 'first_name', 'last_name')[:2]
        self.assertEqual(len(values), 2)
        self.assertIn('first_name', values[0])
        self.assertNotIn('attributes', values[0])
        self.assertEqual(len(values[0]), 3)
        if default_is_sf:
            self.assertRegex(cast(str, values[0]['pk']), '^003')

        values_list = Contact.objects.values_list('pk', 'first_name', 'last_name')[:2]
        self.assertEqual(len(values_list), 2)
        v0 = values[0]
        self.assertEqual(values_list[0], (v0['pk'], v0['first_name'], v0['last_name']))

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_double_delete(self) -> None:
        """Test that repeated delete of the same object is ignored

        the same way like "DELETE FROM Contact WHERE Id='deleted yet'" would do.
        """
        contact = Contact(last_name='sf_test',
                          owner=User.objects.get(Username=current_user))
        contact.save()
        contact_id = contact.pk
        Contact(pk=contact_id).delete()

        # Id of a deleted object or a too small valid Id shouldn't raise
        with warnings.catch_warnings(record=True) as w:
            Contact(pk=contact_id).delete()
            self.assertIs(w[-1].category, SalesforceWarning)
            self.assertIn('ENTITY_IS_DELETED', str(w[-1].message))
        # Simulate the same with obsoleted oauth session
        # It is not possible to use salesforce.auth.expire_token() to simulate
        # expiration because it forces reauhentication before the next request
        salesforce.auth.oauth_data[sf_alias]['access_token'] = 'something invalid/expired'
        with self.assertWarns(SalesforceWarning) as cm:
            Contact(pk=contact_id).delete()
            message = cm.warnings[0].message
            assert isinstance(message, SalesforceWarning)
            self.assertIn('ENTITY_IS_DELETED', message.args[0])
        with self.assertWarns(SalesforceWarning) as cm:
            # Id of completely deleted item or fake but valid item.
            Contact(pk='003000000000000AAA').delete()
        # bad_id = '003000000000000AAB' # Id with an incorrect uppercase mask
        # self.assertRaises(salesforce.backend.base.SalesforceError, Contact(pk=bad_id).delete)

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    @skipUnless(len(sf_databases) > 1, "Only one SF database found.")
    def test_multiple_sf_databases(self) -> None:
        """Test a connection to two sf sandboxes of the same organization.
        """
        other_db = [db for db in sf_databases if db != sf_alias][0]
        c1 = Contact(last_name='sf_test 1')
        c2 = Contact(last_name='sf_test 2')
        c1.save()
        c2.save(using=other_db)
        try:
            user1 = refresh(c1).owner
            user2 = refresh(c2).owner
            username1 = user1.Username
            username2 = user2.Username
            # Verify different usernames, like it is usual in sandboxes
            self.assertNotEqual(user1._state.db, user2._state.db)
            self.assertNotEqual(username1, username2)
            expected_user2 = connections[other_db].settings_dict['USER']
            self.assertEqual(username1, current_user)
            self.assertEqual(username2, expected_user2)
        finally:
            c1.delete()
            c2.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_expired_auth_id(self) -> None:
        """Test the code for expired auth ID for multiple SF databases.

        No similar test exists for a single db.
        """
        self.assertGreaterEqual(len(sf_databases), 1)
        objects = []
        for db in sf_databases:
            c = Contact(last_name='sf_test %s' % db)
            c.save(using=db)
            objects.append(c)
        try:
            # simulate that a request with invalid/expired auth ID re-authenticates
            # and succeeds.
            for db in sf_databases:
                salesforce.auth.oauth_data[db]['access_token'] += 'simulated invalid/expired'
            for x in objects:
                self.assertTrue(refresh(x))
        finally:
            for x in objects:
                x.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_only_fields(self) -> None:
        """Verify that access to "only" fields doesn't require a request, but others do.
        """
        # print([(x.pk, x.last_name) for x in Contact.objects.only('last_name').order_by('pk')[:2]])
        with self.lazy_assert_n_requests(0):
            sql = User.objects.only('Username').query.get_compiler('salesforce').as_sql()[0]
            self.assertEqual(sql, 'SELECT User.Id, User.Username FROM User')

        with self.lazy_assert_n_requests(1):
            user = User.objects.only('Username').order_by('pk')[0]

        # Verify that deferred fields work
        with self.lazy_assert_n_requests(0):
            user.pk
            user_username = user.Username
        self.assertIn('@', user_username)

        with self.lazy_assert_n_requests(1):
            self.assertGreater(len(user.LastName), 0)

        with self.lazy_assert_n_requests(1):
            self.assertEqual(user_username, User.objects.get(pk=user.pk).Username)

        with self.lazy_assert_n_requests(1):
            xx = Contact.objects.only('last_name').order_by('pk')[0]

        with self.lazy_assert_n_requests(0):
            xx.last_name

        with self.lazy_assert_n_requests(1):
            xy = Contact.objects.only('account').order_by('pk')[1]

        with self.lazy_assert_n_requests(0):
            xy.account_id

        with self.lazy_assert_n_requests(1):
            self.assertGreater(len(xy.last_name), 0)

        with self.lazy_assert_n_requests(0):
            xy.last_name

        self.lazy_check()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_defer_fields(self) -> None:
        """Verify that access to a deferred field requires a new request, but others don't.
        """
        with self.lazy_assert_n_requests(1):
            contact = Contact.objects.defer('email')[0]
        with self.lazy_assert_n_requests(0):
            _ = contact.last_name
        with self.lazy_assert_n_requests(1):
            _ = contact.email
        self.lazy_check()
        _  # noqa

    def test_incomplete_raw(self) -> None:
        """Test that omitted model fields can be queried by dot."""
        with self.lazy_assert_n_requests(1):
            ret = list(Contact.objects.raw("select Id from Contact where FirstName != '' limit 2000"))
        with self.lazy_assert_n_requests(1):
            last_name = ret[0].last_name
        self.assertTrue(last_name and 'last' not in last_name.lower())
        with self.lazy_assert_n_requests(1):
            first_name = ret[0].first_name
        with self.lazy_assert_n_requests(0):
            last_name = ret[0].last_name
        assert first_name
        self.lazy_check()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_filter_by_more_fk_to_the_same_model(self) -> None:
        """Test a filter with more relations to the same model.

        Verify that aliases are correctly decoded in the query compiler.
        """
        test_lead = Lead(Company='sf_test lead', LastName='name')
        test_lead.save()
        try:
            qs = Lead.objects.filter(pk=test_lead.pk,
                                     owner__Username=current_user,
                                     last_modified_by__Username=current_user)
            sql, params = qs.query.get_compiler('salesforce').as_sql()
            sql_parts = set(sql.split(' WHERE (')[1].rstrip(')').split(' AND '))
            self.assertEqual(sql_parts,
                             {'Lead.Id = %s', 'Lead.Owner.Username = %s', 'Lead.LastModifiedBy.Username = %s'})
            refreshed_lead = qs.get()
            self.assertEqual(refreshed_lead.pk, test_lead.pk)
        finally:
            test_lead.delete()

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_filter_by_more_fk_to_the_same_model_subquery(self) -> None:
        """Test a filter with more relations to the same model.

        Verify that aliases are correctly decoded in the query compiler.
        """
        test_lead = Lead(Company='sf_test lead', LastName='name')
        test_lead.save()
        test_task = Task(who=test_lead)
        test_task.save()
        try:
            qs = Task.objects.filter(
                who__in=Lead.objects.filter(
                    pk=test_lead.pk,
                    owner__Username=current_user,
                    last_modified_by__Username=current_user)
            )
            sql, params = qs.query.get_compiler('salesforce').as_sql()
            self.assertRegex(sql, r'SELECT Task.Id, .* FROM Task WHERE Task.WhoId IN \(SELECT ')
            self.assertIn('Lead.Owner.Username = %s', sql)
            self.assertIn('Lead.LastModifiedBy.Username = %s', sql)
            refreshed_lead = qs[:1]
            self.assertEqual(len(refreshed_lead), 1)
        finally:
            test_task.delete()
            test_lead.delete()

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_many2many_relationship(self) -> None:
        """Verify that the related set of Many2Many relationship works

        Test for issue #55
        """
        contact = Contact.objects.all()[0]
        oppo = Opportunity(name='test op', stage='Prospecting', close_date=datetime.date.today())
        oppo.save()
        oc = OpportunityContactRole(opportunity=oppo, contact=contact, role='sponsor')
        oc.save()
        oc2 = OpportunityContactRole(opportunity=oppo, contact=contact, role='evaluator')
        oc2.save()
        try:
            qs = contact.opportunities.all()
            sql, params = qs.query.get_compiler('salesforce').as_sql()
            self.assertRegex(sql,
                             r'SELECT .*OpportunityContactRole\.Opportunity\.StageName.* '
                             'FROM OpportunityContactRole WHERE OpportunityContactRole.ContactId =')
            self.assertEqual([x.pk for x in qs], 2 * [oppo.pk])
        finally:
            oc2.delete()
            oc.delete()
            oppo.delete()

    def test_many2many_relationship_filter(self) -> None:
        """Verify that ManyToMany relationship can be filtered by a condition on remote object

        Test for PR #205
        """
        contact = Contact.objects.all()[0]
        oppo = Opportunity(name='test op', stage='Prospecting', close_date=datetime.date.today())
        oppo.save()
        oc = OpportunityContactRole(opportunity=oppo, contact=contact, role='sponsor')
        oc.save()
        oc2 = OpportunityContactRole(opportunity=oppo, contact=contact, role='evaluator')
        oc2.save()
        try:
            qs = Contact.objects.filter(opportunity_roles__opportunity__name='test op')
            self.assertEqual(list(qs), 2 * [contact])
        finally:
            oc2.delete()
            oc.delete()
            oppo.delete()

    def test_filter_custom(self) -> None:
        """Verify that relations between custom and builtin objects

        are correctly compiled. (__r, __c etc.)
        """
        if 'django_Test__c' not in sf_tables():
            self.skipTest("Not found custom object 'django_Test__c'")
        qs = Attachment.objects.filter(parent__name='abc')
        # "SELECT Attachment.Id FROM Attachment WHERE Attachment.Parent.Name = 'abc'"
        list(qs)
        qs2 = Test.objects.filter(contact__last_name='Johnson')
        # "SELECT ... FROM django_Test__c WHERE django_Test__c.Contact__r.LastName = 'Johnson'")
        list(qs2)
        qs = Attachment.objects.filter(parent__in=Test.objects.filter(contact__last_name='Johnson'))
        list(qs)
        # test the '__range' lookup
        qs = Test.objects.filter(contact__name__range=('a', 'b')).values('pk').sf(minimal_aliases=True)
        expect_sql = "SELECT Id FROM django_Test__c WHERE (Contact__r.Name >= 'a' AND Contact__r.Name <= 'b')"
        list(qs)
        self.assertEqual(str(qs.query), expect_sql)

    def test_using_none(self) -> None:
        alias = getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce')
        self.assertEqual(Contact.objects.using(None)._db, alias)

    def test_with_offst(self) -> None:
        list(Contact.objects.all()[1:2])
        list(Contact.objects.all()[2000:])

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_dynamic_fields(self) -> None:
        """Test that fields can be copied dynamically from other model"""
        self.assertTrue(models_template)
        self.assertIn('@', Organization.objects.get().created_by.Username)

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    @override_settings(SALESFORCE_QUERY_TIMEOUT=(4, 30))
    def test_big_soql(self) -> None:
        """Test that a query of length almost 100000 is possible"""
        contact = Contact.objects.all()[0]
        # 4534 items * 22 characters == 99748
        names = [get_random_string(18) for _ in range(4534)]
        names.append(contact.name)
        qs = Contact.objects.filter(name__in=names)
        with PatchedSfConnection(sf_alias, verbs=['use_debug_info']):
            self.assertIn(contact.pk, list(qs.values_list('pk', flat=True)))
        processed_soql = connections[sf_alias].connection.debug_info['soql'][2]
        self.assertGreater(len(processed_soql), 90000)

    def test_empty_slice(self) -> None:
        """Test queryset with empty slice - if high/low limits equals"""
        self.assertEqual(len(Contact.objects.all()[1:1]), 0)

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def test_select_explan(self) -> None:
        """Test EXPLAIN SELECT ..."""
        qs = Contact.objects.all()[:2]
        self.assertRegex(qs.explain(), r"^{'plans': \[{")

    def test_api_usage(self) -> None:
        """Test that API usage is updated by executing requests"""
        api_usage: ApiUsage = connections[sf_alias].connection.api_usage
        self.assertGreater(api_usage.api_usage, 0)
        self.assertGreater(api_usage.api_limit, 5000)

    def test_extract_date(self) -> None:
        campaign = Campaign.objects.create(name='test something', number_sent=3)
        try:
            # group datetime by year
            list(Campaign.objects.values('created_date__year').annotate(total=Sum('number_sent')))
            # filter datetime last year
            list(Campaign.objects.filter(created_date__year=2023, number_sent__gte=100))
            # filter date last year and group by month
            list(Campaign.objects.filter(start_date__year=2023).values('start_date__month').annotate(cnt=Count('pk')))
            # filter datetatetime by month
            list(Campaign.objects.filter(created_date__month=1))
            # filter datetatetime by quarter
            list(Campaign.objects.filter(created_date__quarter=1))
            # group by date from a DateTime field
            list(Campaign.objects.values(c_date=functions.TruncDate('created_date')).annotate(cnt=Count('start_date')))
            # example group by all possible type of datetime lookup expressions
            list(Campaign.objects.values('start_date__year', 'start_date__month', 'start_date__day',
                                         'start_date__week', 'start_date__week_day',
                                         'start_date__quarter').annotate(total=Count('pk')))
            list(Campaign.objects.values('created_date__year', 'created_date__month', 'created_date__day',
                                         'created_date__week', 'created_date__week_day',
                                         'created_date__hour', 'created_date__quarter').annotate(total=Count('pk')))
        finally:
            campaign.delete()


# ============= Tests that need setUp Lead ==================


class BasicLeadSOQLTest(TestCase):
    """Tests that use a test Lead"""
    databases = '__all__'

    def setUp(self) -> None:
        """Create our test lead record.
        """
        def add_obj(obj):
            obj.save()
            self.objs.append(obj)
        #
        self.test_lead = Lead(
            FirstName="User" + uid,
            LastName="Unittest General",
            Email=test_email,
            Status='Open',
            Company="Some company, Ltd.",
        )
        self.objs: List[SalesforceModel] = []
        self.test_lead.save()
        # This is only for demonstration that some test can be run even with
        # non SFDC database SALESFORCE_DB_ALIAS, even if the test expects some
        # contacts and the current user, but most of tests can pass only on SFDC.
        if not default_is_sf:
            add_obj(Contact(last_name='Test contact 1'))
            add_obj(Contact(last_name='Test contact 2'))
            add_obj(User(Username=current_user))

    def tearDown(self) -> None:
        """Clean up our test records.
        """
        if self.test_lead.pk is not None:
            self.test_lead.delete()
        for obj in self.objs:
            if obj.pk is not None:
                obj.delete()
        self.objs = []

    def test_exclude_query_construction(self) -> None:
        """Test that exclude query construction returns valid SOQL.
        """
        contacts = (Contact.objects.filter(first_name__isnull=False)
                    .exclude(email="steve@apple.com", last_name="Wozniak")
                    .exclude(last_name="smith"))
        number_of_contacts = contacts.count()
        self.assertIsInstance(number_of_contacts, int)
        # the default self.test_lead shouldn't be excluded by only one nondition
        leads = (Lead.objects.exclude(Email="steve@apple.com", LastName="Unittest General")
                 .filter(FirstName="User" + uid, LastName="Unittest General"))
        self.assertEqual(leads.count(), 1)

    def test_get(self) -> None:
        """Get the test lead record.
        """
        lead = Lead.objects.get(Email=test_email)
        self.assertEqual(lead.FirstName, 'User' + uid)
        self.assertEqual(lead.LastName, 'Unittest General')
        if not default_is_sf:
            self.skipTest("Default database should be any Salesforce.")
        # test a read only field (formula of full name)
        self.assertEqual(lead.Name, 'User%s Unittest General' % uid)

    def test_not_null(self) -> None:
        """Get the test lead record by isnull condition.
        """
        lead = Lead.objects.get(Email__isnull=False, FirstName='User' + uid)
        self.assertEqual(lead.FirstName, 'User' + uid)
        self.assertEqual(lead.LastName, 'Unittest General')

    def test_update(self) -> None:
        """Update the test lead record.
        """
        test_lead = Lead.objects.get(Email=test_email)
        self.assertEqual(test_lead.FirstName, 'User' + uid)
        test_lead.FirstName = 'Tested'
        test_lead.save()
        self.assertEqual(refresh(test_lead).FirstName, 'Tested')

    def test_save_update_fields(self) -> None:
        """Test the save method with parameter `update_fields`

        that updates only required fields.
        """
        company_orig = self.test_lead.Company
        self.test_lead.Company = 'nonsense'
        self.test_lead.FirstName = 'John'
        self.test_lead.save(update_fields=['FirstName'])
        test_lead = refresh(self.test_lead)
        self.assertEqual(test_lead.FirstName, 'John')
        self.assertEqual(test_lead.Company, company_orig)

    def test_query_all_deleted(self) -> None:
        """Test query for deleted objects (queryAll resource).
        """
        self.test_lead.delete()
        # TODO optimize counting because this can load thousands of records
        count_deleted = Lead.objects.db_manager(sf_alias).query_all(
                ).filter(IsDeleted=True, LastName="Unittest General").count()
        if not default_is_sf:
            self.skipTest("Default database should be any Salesforce.")
        self.assertGreaterEqual(count_deleted, 1)
        count_deleted2 = Lead.objects.filter(IsDeleted=True, LastName="Unittest General").query_all().count()
        self.assertGreaterEqual(count_deleted2, count_deleted)

        self.test_lead.pk = None
        self.test_lead.save()  # save anything again to be cleaned finally

    @skipUnless(default_is_sf, "Default database should be any Salesforce.")
    def test_generic_type_field(self) -> None:
        """Test that a generic foreign key can be filtered by type name and
        the type name can be referenced.
        """
        test_contact = Contact(first_name='sf_test', last_name='my')
        test_contact.save()
        note_1 = Note(title='note for Lead', parent_id=self.test_lead.pk)
        note_2 = Note(title='note for Contact', parent_id=test_contact.pk)
        note_1.save()
        note_2.save()
        try:
            self.assertEqual(Note.objects.filter(parent_type='Contact')[0].parent_type, 'Contact')
            self.assertEqual(Note.objects.filter(parent_type='Lead')[0].parent_type, 'Lead')

            note = Note.objects.filter(parent_type='Contact')[0]
            parent_model = getattr(salesforce.testrunner.example.models, note.parent_type)
            parent_object = parent_model.objects.get(pk=note.parent_id)
            self.assertEqual(parent_object.pk, note.parent_id)
        finally:
            note_1.delete()
            note_2.delete()
            test_contact.delete()



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

import logging

from django import shortcuts

from salesforce.testrunner.example import models, forms

log = logging.getLogger(__name__)


def list_accounts(request):
    accounts = models.Account.objects.all()[0:5]

    return shortcuts.render(request, 'list-accounts.html',
                            dict(title="List First 5 Accounts", accounts=accounts)
                            )


def search_accounts(request):
    accounts = []
    if request.method == 'POST':
        form = forms.SearchForm(request.POST)
        if form.is_valid():
            accounts = models.Account.objects.filter(Name__icontains=form.cleaned_data['query'])
    else:
        form = forms.SearchForm()

    return shortcuts.render(request, 'search-accounts.html',
                            dict(title="Search Accounts by Email", accounts=accounts, form=form)
                            )



"""Test of methods that can modify more records by one request
or can combine read and write and read to one request.
"""

from typing import List, Optional, Tuple
from django.conf import settings
from django.test import TestCase
from salesforce import SalesforceError
from salesforce.backend.test_helpers import default_is_sf, skipUnless, expectedFailure
from salesforce.testrunner.example.models import Campaign, CampaignMember, Contact

SF_EXAMPLE_CUSTOM_INSTALLED = getattr(settings, 'SF_EXAMPLE_CUSTOM_INSTALLED', False)


def common_setup_contacts(n: int) -> List[Contact]:
    contact_names = ['sf_test {}'.format(i) for i in range(n)]
    contacts = {x.name: x for x in Contact.objects.filter(name__startswith='sf_test ')
                if x.name in contact_names}
    missing_contacts = [Contact(last_name=x) for x in contact_names if x not in contacts]
    if missing_contacts:
        Contact.objects.bulk_create(missing_contacts)
    return list(contacts.values()) + missing_contacts


class BulkUpdateTest(TestCase):
    """The method queryset.bulk_update() is tested by a Contact with a unique custom field"""

    databases = '__all__'
    contacts = None  # type: List[Contact]

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.contacts = common_setup_contacts(9)
        if SF_EXAMPLE_CUSTOM_INSTALLED and not Contact.objects.update(vs=None) >= 9:
            Contact.objects.bulk_create([Contact(last_name='sf_test {}'.format(i)) for i in range(9)])

    @skipUnless(SF_EXAMPLE_CUSTOM_INSTALLED, "requires Salesforce customization")
    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        Contact.objects.update(vs=None)

    def test_bulk_update_normal(self) -> None:
        contacts = Contact.objects.all()[:9]
        for i, x in enumerate(contacts):
            x.vs = i
        Contact.objects.bulk_update(contacts, ['vs'])
        qs = Contact.objects.filter(vs__gte=0)
        self.assertEqual(sorted(qs.values_list('vs', flat=True)), [0, 1, 2, 3, 4, 5, 6, 7, 8])

    def common_bulk_update_error(self, expected_count: int, all_or_none: Optional[bool] = None) -> None:
        contacts = list(Contact.objects.all()[:9])
        for i, x in enumerate(contacts):
            x.vs = i
        contacts[6].vs = 5  # duplicate value 5
        with self.assertRaises(SalesforceError) as cm:
            Contact.objects.bulk_update(contacts, ['vs'], batch_size=4,  # type: ignore[call-arg] # all_or_none
                                        all_or_none=all_or_none)
        self.assertIn('Contact  DUPLICATE_VALUE', cm.exception.args[0])
        self.assertEqual(Contact.objects.exclude(vs=None).count(), expected_count)

    def test_bulk_update_error(self) -> None:
        """Verify that also valid data in the chunk after an error are updated"""
        self.common_bulk_update_error(expected_count=4 + 3 + 0)

    def test_bulk_update_error_all_or_none_true(self) -> None:
        """Verify that no data in the invalid chunk and after it are not updated"""
        self.common_bulk_update_error(expected_count=4 + 0 + 0, all_or_none=True)

    # def test_bulk_update_error_all_or_none_false(self) -> None:
    #     """Verify that all valid data are updated, even in batches after error"""
    #     self.common_bulk_update_error(expected_count=4 + 3 + 1, all_or_none=False)


class QuerysetUpdateTest(TestCase):
    """The method queryset.update() is tested by a Contact with a unique custom field"""

    databases = '__all__'
    contacts = None  # type: List[Contact]

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.contacts = common_setup_contacts(2)
        if SF_EXAMPLE_CUSTOM_INSTALLED:
            assert Contact.objects.update(vs=None) >= 2

    @skipUnless(SF_EXAMPLE_CUSTOM_INSTALLED, "requires Salesforce customization")
    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        Contact.objects.update(vs=None)

    def common_qs_update_error(self, expected_count: int, all_or_none: Optional[bool] = None) -> None:
        with self.assertRaises(SalesforceError) as cm:
            Contact.objects.sf(all_or_none=all_or_none).update(vs=1)  # type: ignore[attr-defined] # sf
        self.assertIn('Contact  DUPLICATE_VALUE', cm.exception.args[0])
        self.assertEqual(Contact.objects.filter(vs=1).count(), expected_count)

    def test_qs_update_error(self) -> None:
        """Verify that also valid data in the chunk after an error are updated"""
        self.common_qs_update_error(expected_count=1)

    def test_qs_update_error_all_or_none_true(self) -> None:
        """Verify that no data in the invalid chunk and after it are not updated"""
        self.common_qs_update_error(expected_count=0, all_or_none=True)

    # the parameter `all_or_none=False` can not be easily tested on Salesforce
    # (without special validation rules on Salesforce)


class BulkCreateSimpleTest(TestCase):
    """Simple test of method queryset.bulk_create()"""

    databases = '__all__'

    def test(self) -> None:
        contacts = [Contact(last_name=f"sf_test bulk {i}") for i in range(3)]
        Contact.objects.bulk_create(contacts[:1])
        Contact.objects.bulk_create(contacts[1:])
        contact_ids = [x.pk for x in contacts]
        self.assertEqual(Contact.objects.filter(pk__in=contact_ids).count(), 3)
        Contact.objects.filter(pk__in=contact_ids).delete()


class BulkCreateTest(TestCase):
    """The method queryset.bulk_create() is tested by a CampaignMember"""

    databases = '__all__'
    contacts = []  # type: List[Contact]
    campaign = None  # type: Campaign

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.contacts = common_setup_contacts(4)
        campaigns = Campaign.objects.all()[:1]
        if campaigns:
            cls.campaign = campaigns[0]
        else:
            cls.campaign = Campaign.objects.create(name='test campaign')
        CampaignMember.objects.filter(campaign=cls.campaign, contact__in=cls.contacts).delete()

    def tearDown(self) -> None:
        CampaignMember.objects.filter(campaign=self.campaign, contact__in=self.contacts).delete()

    @skipUnless(default_is_sf, "depends on Salesforce database.")
    def common_bulk_create_error(self,
                                 data: Tuple[int, ...],
                                 expected: Tuple[int, int, int, int],
                                 all_or_none: Optional[bool] = None) -> None:
        expect_errors, expect_rollbacks, expect_success, expect_n = expected
        members = [
            CampaignMember(campaign=self.campaign, contact=self.contacts[i], status='Sent')
            for i in data
        ]
        qs = CampaignMember.objects
        if all_or_none is not None:
            qs = qs.sf(all_or_none=all_or_none)  # type: ignore[attr-defined] # noqa # sf
        with self.assertRaises(SalesforceError) as cm:
            qs.bulk_create(members, batch_size=3, ignore_conflicts=True)
        error_summary = [x.strip() for x in cm.exception.args[0].split('\n') if 'Error Summary' in x][0]
        expected_summary = 'Error Summary: errors={}, rollback/cancel={}, success={}'.format(
            expect_errors, expect_rollbacks, expect_success)
        self.assertEqual(error_summary, expected_summary)
        self.assertIn('CampaignMember  DUPLICATE_VALUE', cm.exception.args[0])
        qs = CampaignMember.objects.filter(campaign=self.campaign, contact__in=self.contacts)
        self.assertEqual(qs.count(), expect_n)
        inserted_n = len([x for x in members if x.pk is not None])
        self.assertEqual(inserted_n, expect_n)

    @expectedFailure
    def test_bulk_create_error(self) -> None:
        """Verify that also valid data in the chunk after an error are inserted"""
        self.common_bulk_create_error((0, 1, 2, 3, 3, 3), expected=(2, 0, 1, 4))

    @expectedFailure
    def test_bulk_create_error_all_or_none_true(self) -> None:
        """Verify that no data in the invalid chunk and after it are not inserted"""
        self.common_bulk_create_error((0, 1, 2, 3, 3, 3), expected=(2, 1, 0, 3), all_or_none=True)

    # def test_bulk_create_error_all_or_none_false(self) -> None:
    #     """Verify that all valid data are inserted, even in batches after error"""
    #     self.common_bulk_create_error(expected=(2, 0, 1), all_or_none=False)



from django.urls import include, path

from django.contrib import admin
admin.autodiscover()

urlpatterns = [
    path(r'', include('salesforce.testrunner.example.urls')),
    path(r'admin/doc/', include('django.contrib.admindocs.urls')),
    path(r'admin/', admin.site.urls),
]



"""
Module for models for a combined storage with a Salesforce and normal database.

usage purposes e.g.:
  - Backup salesforce objects including its primary keys.
  - Use the backup in queries, including ForeignKeys related correctly to the same database.
  - Update the object in Salesfoce to the original values from backup object
    if the primary key still exists in salesforce.
  - Create a new non-salesforce object with an automatic uuid pk.
  - Use it as a much faster alternative than "data sandbox refresh".

The non-salesforce database uses a default uuid() key or a provided salesforce Id value.

The module less supported in future Django versions or unsupported.

Interesting not yet implemented ideas are:
  - Use a normal Salesforce backup zip like a read-only database (list
    of Salesforce objects that can be filtered and fast saved to a non-salesforce database
    by bulk_create().
  - Could be useful for tests with realistic data on a database with rollback.
"""

from typing import Any, Generic, TYPE_CHECKING
from django.db import models, router
from django.db.backends.base.base import BaseDatabaseWrapper as DatabaseWrapper

from salesforce.backend import DJANGO_30_PLUS
from salesforce.backend.indep import get_sf_alt_pk
from salesforce.models import *  # NOQA; pylint:disable=wildcard-import,unused-wildcard-import
from salesforce.models import SalesforceAutoField, SalesforceModelBase, SF_PK, _T
from salesforce.router import is_sf_database
if not TYPE_CHECKING:
    from salesforce.backend import manager

__all__ = ('SalesforceModel',)


class SfCharAutoField(SalesforceAutoField):
    """Auto field that allows Salesforce ID or UUID in an alternate database"""

    # db_returning = False  # this was a simple fix for Django >= 3.0,
    #                       # but a fix by "_do_insert()" is better.

    validators = ()  # type: ignore[assignment] # easier than an immutable list

    def get_internal_type(self) -> str:
        return 'CharField'

    def db_type(self, connection: DatabaseWrapper) -> str:
        if connection.vendor != 'salesforce':
            # it is 'varchar(32)'
            return models.CharField(max_length=32).db_type(connection=connection)
        return 'AutoField'

    def rel_db_type(self, connection: DatabaseWrapper) -> str:
        if connection.vendor != 'salesforce':
            return models.CharField(max_length=32).db_type(connection=connection)
        return 'AutoField'


if TYPE_CHECKING:
    class SalesforceModel(models.Model, Generic[_T],  # type: ignore[no-redef] # pylint:disable=function-redefined
                          metaclass=SalesforceModelBase):
        _salesforce_object = ...

        def __init__(self, *args: Any, **kwargs: Any) -> None:  # pylint:disable=super-init-not-called
            tmp = models.manager.Manager()  # type: models.manager.Manager[_T]
            self.objects = tmp
else:
    # pylint:disable=function-redefined
    class SalesforceModel(models.Model, metaclass=SalesforceModelBase):  # pylint:disable=function-redefined
        """
        Abstract model class for Salesforce objects that can be saved to other db.

        (It is not a subclass of salesforce.models.SalesforceModel. That is not
        a big problem if we don't check inheritance but only the '_salesforce_object'
        attribute or if we use only this or only the original implementation.)
        """
        _salesforce_object = 'extended'
        objects = manager.SalesforceManager()  # type: manager.SalesforceManager[_T]

        class Meta:
            # pylint:disable=duplicate-code
            abstract = True
            base_manager_name = 'objects'

        id = SfCharAutoField(primary_key=True, name=SF_PK, db_column='Id', verbose_name='ID', auto_created=True,
                             editable=False)

        def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
            using = using or router.db_for_write(self.__class__, instance=self)
            if self.pk is None and not force_update and not is_sf_database(using):
                self.pk = get_sf_alt_pk()
            super().save(force_insert=force_insert, force_update=force_update,
                         using=using, update_fields=update_fields)
            if not isinstance(self.pk, str):
                raise ValueError("The primary key value is not assigned correctly")

        if DJANGO_30_PLUS:

            def _do_insert(self, manager, using, fields, returning_fields, raw):  # pylint:disable=redefined-outer-name
                # the check "is_sf_database(using)" is used for something unexpected
                if self.pk and not is_sf_database(using):
                    returning_fields = []
                return super()._do_insert(manager, using, fields, returning_fields, raw)



"""
Tests for `salesforce.utils`
"""
# pylint:disable=protected-access

from unittest import skipUnless

from django.test import TestCase

from salesforce.dbapi.driver import beatbox
from salesforce.testrunner.example.models import Account, Lead
from salesforce.utils import convert_lead


class UtilitiesTest(TestCase):
    databases = '__all__'

    @skipUnless(beatbox, "Beatbox needs to be installed in order to run this test.")
    def test_lead_conversion(self):
        """
        Create a Lead object within Salesforce and try to
        convert it, convert/merge it with the information from a duplicit Lead,
        then clean all the generated objects.
        """
        lead = Lead(FirstName="Foo", LastName="Bar", Company="django-salesforce",
                    Street='Test Avenue 45')
        lead.save()
        lead2 = Lead(FirstName="Foo", LastName="Bar", Company="django-salesforce",
                     Phone='123456789')
        lead2.save()
        ret = None
        try:
            # convert the first Lead
            ret = convert_lead(lead, doNotCreateOpportunity=True)
            # print("Response from convertLead: " +
            #       ', '.join('%s: %s' % (k, v) for k, v in sorted(ret.items())))
            expected_names = set(('accountId', 'contactId', 'leadId', 'opportunityId', 'success'))
            # The field 'relatedPersonAccountId' is present in the instances ver. 51.0+ Spring '21
            self.assertEqual(set(ret).difference(['relatedPersonAccountId']), expected_names)
            self.assertEqual(ret['success'], 'true')
            # merge the new Account with the second Lead
            ret2 = convert_lead(lead2, doNotCreateOpportunity=True, accountId=ret['accountId'])
            account = Account.objects.get(pk=ret['accountId'])
            # verify that account is merged
            self.assertEqual(ret2['accountId'], account.pk)
            self.assertEqual(account.BillingStreet, 'Test Avenue 45')
            self.assertEqual(account.Phone, '123456789')
        finally:
            # Cleaning up...
            if ret:
                # Deleting the Account object will also delete the related Contact
                # and Opportunity objects.
                try:
                    account = Account.objects.get(pk=ret['accountId'])
                except Exception:  # pylint:disable=broad-except        # pragma: no cover
                    # this allows to recycle the account even if the queryset code is broken
                    account = Account(pk=ret['accountId'])
                    account._state.db = lead._state.db
                account.delete()
            lead.delete()   # FYI, ret['leadId'] == lead.pk
            lead2.delete()



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Automatic table creation is not supported by the Salesforce backend. (like django.db.backends.*.creation)
"""
import logging

from django.db.backends.base.creation import BaseDatabaseCreation

log = logging.getLogger(__name__)


class DatabaseCreation(BaseDatabaseCreation):
    # pylint:disable=abstract-method  # undefined '_clone_test_db'

    def create_test_db(self, verbosity=1, autoclobber=False, serialize=True, keepdb=False):
        test_database_name = self._get_test_db_name()

        if verbosity >= 1:
            test_db_repr = ''
            if verbosity >= 2:
                test_db_repr = " ('%s')" % test_database_name
            log.info("Ignoring test database creation for alias '%s'%s...",
                     self.connection.alias, test_db_repr)

        return test_database_name

    def destroy_test_db(self, old_database_name=None, verbosity=1, keepdb=False, suffix=None):
        test_database_name = self.connection.settings_dict['NAME']
        if verbosity >= 1:
            test_db_repr = ''
            if verbosity >= 2:
                test_db_repr = " ('%s')" % test_database_name
            log.info("No test database to destroy for alias '%s'%s...",
                     self.connection.alias, test_db_repr)
        self.connection.settings_dict['NAME'] = old_database_name



"""
Common helpers for tests, like test decorators
"""
import sys
import uuid
from unittest import skip as skip, skipUnless as skipUnless, expectedFailure as expectedFailure, TestCase  # NOQA pylint:disable=unused-import
from typing import Union

import django
from django.conf import settings

from salesforce import router
from salesforce.dbapi.test_helpers import (  # NOQA pylint:disable=unused-import
    LazyTestMixin as LazyTestMixin, expectedFailureIf as expectedFailureIf,
    QuietSalesforceErrors as QuietSalesforceErrors)

# uid strings for tests that accidentally run concurrent
uid_random = '-' + str(uuid.uuid4())[:7]
# this is the same as the name of tox test environment, e.g. 'dj30-py38'
uid_version = 'dj{0}{1}-py{2}{3}'.format(*(django.VERSION[:2] + sys.version_info[:2]))

sf_alias = getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce')
default_is_sf = router.is_sf_database(sf_alias)
current_user: str = settings.DATABASES[sf_alias].get('USER', '')


def strtobool(val: Union[str, bool]) -> bool:
    """instead of deprecated distutils.util.strtobool(...)"""
    return str(val).lower() in ("y", "yes", "t", "true", "on", "1")



from unittest import TestCase
from salesforce.dbapi.subselect import (
    find_closing_parenthesis, split_subquery, transform_except_subquery,
    mark_quoted_strings, subst_quoted_strings, simplify_expression,
)


class TestSubSelectSearch(TestCase):
    def test_parenthesis(self):
        self.assertEqual(find_closing_parenthesis('() (() (())) ()', 0), (0, 2))
        self.assertEqual(find_closing_parenthesis('() (() (())) ()', 2), (3, 12))
        self.assertEqual(find_closing_parenthesis('() (() (())) ()', 3), (3, 12))
        self.assertEqual(find_closing_parenthesis('() (() (())) ()', 6), (7, 11))
        self.assertEqual(find_closing_parenthesis('() (() (())) ()', 13), (13, 15))
        self.assertRaises(AssertionError, find_closing_parenthesis, '() (() (())) ()', 1)

    def test_subquery(self):
        def func(x):  # pylint:disable=unused-argument
            return '*transformed*'

        sql = "SELECT a, (SELECT x FROM y) FROM b WHERE (c IN (SELECT p FROM q WHERE r = %s) AND c = %s)"
        expected = "*transformed*(SELECT x FROM y)*transformed*(SELECT p FROM q WHERE r = %s)*transformed*"
        self.assertEqual(transform_except_subquery(sql, func), expected)

    def test_split_subquery(self):
        sql = " SELECT a, ( SELECT x FROM y) FROM b WHERE (c IN (SELECT p FROM q WHERE r = %s) AND c = %s)"
        expected = ("SELECT a, (&) FROM b WHERE (c IN (&) AND c=%s)",
                    [("SELECT x FROM y", []),
                     ("SELECT p FROM q WHERE r=%s", [])
                     ])
        self.assertEqual(split_subquery(sql), expected)

    def test_nested_subquery(self):
        def func(x):  # pylint:disable=unused-argument
            return '*transformed*'

        sql = "SELECT a, (SELECT x, (SELECT p FROM q) FROM y) FROM b"
        expected = "*transformed*(SELECT x, (SELECT p FROM q) FROM y)*transformed*"
        self.assertEqual(transform_except_subquery(sql, func), expected)

    def test_split_nested_subquery(self):
        sql = "SELECT a, (SELECT x, (SELECT p FROM q) FROM y) FROM b"
        expected = ("SELECT a, (&) FROM b",
                    [("SELECT x, (&) FROM y",
                      [("SELECT p FROM q", [])]
                      )]
                    )
        self.assertEqual(split_subquery(sql), expected)


class ReplaceQuotedStringsTest(TestCase):
    def test_subst_quoted_strings(self):
        def inner(sql, expected):
            result = mark_quoted_strings(sql)
            self.assertEqual(result, expected)
            self.assertEqual(subst_quoted_strings(*result), sql)
        inner("where x=''", ("where x=@", ['']))
        inner("a'bc'd", ("a@d", ['bc']))
        inner(r"a'bc\\'d", ("a@d", ['bc\\']))
        inner(r"a'\'\\'b''''", ("a@b@@", ['\'\\', '', '']))
        self.assertRaises(AssertionError, mark_quoted_strings, r"a'bc'\\d")
        self.assertRaises(AssertionError, mark_quoted_strings, "a'bc''d")

    def test_simplify_expression(self):
        self.assertEqual(simplify_expression(' a \t b  c . . d '), 'a b c..d')



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

import base64
import re
from hashlib import sha256
from typing import Optional
import django
from django.conf import settings
from salesforce.backend import max_django
from salesforce.dbapi.exceptions import LicenseError


def check_enterprise_license(  # pylint:disable=too-many-locals
        msg: Optional[str] = None, required: int = 1, key: str = '', lts: bool = False
        ) -> None:
    """Check that a valid license key for the enterprise version exists"""
    key = key or getattr(settings, 'DJSF_LICENSE_KEY', '') or '//'
    text, level_text, signature_text = key.rsplit('/', 2)
    level = int(level_text or 1) if text else 0
    if level < max(required, 1):
        raise LicenseError(msg or "This command requires the enterprise license key")

    # This check is an integer arithmetic puzzle that does not use anything except Python
    # and Django settings. It is evaluated in less than half milisecond on a notebook CPU.
    # There is no cryptography, no network access. The license key never expires.
    # It is a safe code. Please do not discuss publicly about other aspects how the
    # license key is weak or strong. Many easier ways exist how to skip a verification
    # without analyzing this. :-)

    # A legal way to skip the enterprise check is to use django-salesforce-agpl
    # <https://github.com/django-salesforce/django-salesforce-agpl>
    # and accept the restrictive AGPL license that requires you provide all your Django
    # source code where you use django-salesforce available to all users who could
    # use your project by network (by your web app, by your backend app etc.) and
    # publish it under a compatible license.
    # That AGPL license is useful especially for education and for exclusive open
    # source contribution. In other cases, if you want to use the enterprise features,
    # you probably need to sponsor this project to make its development sustainable.

    n, o, q, r, s, t, u, v, w, x = 32, 63, 220, 1, 8, 9223372036854775807, 63, 69, 'utf8', 'big'
    h0 = base64.b64decode(signature_text, b'+.')
    h = int.from_bytes(h0, x)
    b, f, d, e, g, j, k, h, m = 0, 0, n, r, 0, h & t, (h >> o) & u, h >> v, 0
    p = sha256(re.sub("[ &']", "-", text).strip('-').encode(w)).digest() + int.to_bytes(h, s, x)
    for a in range(o):
        p = sha256(p).digest()
        y = int.from_bytes(p, x)
        m, b = max(m, bool(a == k) * y), b ^ (bool(j & (r << a)) * y)
    for a in range(q):
        g, f, d, e = g | (bool(b & e & m) << f), f + bool(m & e), d - bool(j & e), e << r
    z = h >> ((len(h0) << 3) - v - 32)
    z <<= 4 * (z == 3516872)  # fix Agil
    z = (z >> (8 * (2 + (z & 1 << 23 == 0)))) & 0x7fff
    if g != level or level > 3 or d:
        raise LicenseError("The enterprise license key is invalid")
    if (484896 if lts else 484896) & 1 << z:
        raise LicenseError("This license key did not carry over to new packackage versions "
                           "after sponsorship ended.")


def check_license_in_latest_django() -> None:
    # see:
    #   https://github.com/django-salesforce/django-salesforce/wiki/Release-cycle-and-Licenses
    # Django 5.0 is free in Django-salesforce >= 5.1
    # Django 5.1 is free in Django-salesforce >= 5.2
    # Django 5.2 LTS < 5.2.15 will free with Django-salesforce 6.0
    # and all versions of Django 5.2 will be free in Django-salesforce 6.2
    last_django = django.VERSION[:2] == max_django
    protected_lts = (django.VERSION[1] == 2 and django.VERSION[0] == max_django[0] - 1 and
                     (max_django[1] == 1 or max_django[1] == 0 and django.VERSION[2:3] >= [15]))
    is_dev_version = django.VERSION[3:] and re.match('(alpha|beta|rc)', django.VERSION[3])
    if (last_django or protected_lts) and not is_dev_version:
        check_enterprise_license(
            "License key is required for django-saleforce used with the last Django version "
            "or in the first half of life of the last LTS version. (read about dual-license)",
            lts=protected_lts)



import unittest
from salesforce.backend.indep import LazyField


class MyField:

    def __init__(self, a=None, **kwargs):
        self.kwargs = kwargs
        self.a = a

    @property
    def kw(self):
        kw = self.kwargs
        kw['a'] = self.a
        return kw


class TestLazyField(unittest.TestCase):

    def test_positional_arg(self):
        MyLazyField = LazyField(MyField)
        field = MyLazyField(1)
        self.assertEqual(field.kw, {})
        self.assertEqual(field.create().kw, {'a': 1})

    def test_noarg_and_update_later(self):
        MyLazyField = LazyField(MyField)
        self.assertEqual(MyLazyField.klass, MyField)
        field = MyLazyField()
        self.assertTrue(not hasattr(field, 'kwargs'))
        self.assertEqual(field.kw, {})
        self.assertEqual(field.create().kw, {'a': None})
        field.update(a=7)
        self.assertEqual(field.kw, {'a': 7})
        self.assertEqual(field.create().kw, {'a': 7})

    def test_positional_arg_as_kw(self):
        MyLazyField = LazyField(MyField)
        self.assertEqual(MyLazyField.klass, MyField)
        field = MyLazyField(a=2)
        self.assertEqual(field.kw, {})
        self.assertEqual(field.create().kw, {'a': 2})

    def test_kw_args(self):
        MyLazyField = LazyField(MyField)
        field = MyLazyField(b=2, c=3)
        self.assertEqual(field.kw, {'b': 2, 'c': 3})
        self.assertEqual(field.create().kw, {'a': None, 'b': 2, 'c': 3})

    def test_mixed_args(self):
        MyLazyField = LazyField(MyField)
        field = MyLazyField(1, b=4)
        self.assertEqual(field.kw, {'b': 4})
        self.assertEqual(field.create().kw, {'a': 1, 'b': 4})

    def test_mismatch_args(self):
        MyLazyField = LazyField(MyField)
        with self.assertRaises(TypeError):
            MyLazyField(1, a=1)

    def test_counter(self):
        MyLazyField = LazyField(MyField)
        counter_0 = LazyField.counter
        field_1 = MyLazyField()
        self.assertEqual(field_1.counter, counter_0)
        field_2 = MyLazyField()
        self.assertEqual(field_1.counter, counter_0)
        self.assertEqual(field_2.counter, counter_0 + 1)
        self.assertEqual(LazyField.counter, counter_0 + 2)

    def test_separated(self):
        MyLazyField = LazyField(MyField)
        field_1 = MyLazyField(a=1)
        field_2 = MyLazyField(a=2)
        self.assertEqual((field_1.create().a, field_2.create().a), (1, 2))



"""
Simplified Python DB API 2.0 driver for Salesforce - (independent on Django)

https://www.python.org/dev/peps/pep-0249/

A basic SQL SELECT and a small part of PEP 0249 is currently implemented here.
Transactions will be never implemented.
SQL commands INSERT, UPDATE and DELETE could be implemented easily, but it is not
important because the current object implementation in Django is good enough.

Purpose:
Django-Salesforce is splitted to a high level part that depends on Django
(especially on a Django version) and a low level part that depends
on Salesforce API, but not on Django.
It is better for development, maintenance and testing.

This module can run without Django installed, but Django is still main purpose.
Configuration parameters are passed by `settings_dict` the same way as in Django.

The high-level public API is importable directly from this module `salesforce.dbapi`.

A connection can be created by:

    >>> from salesforce.dbapi import connect, get_connection
    >>> settings_dict = {'USER': ..., 'PASSWORD': ..., 'HOST': ...}
    >>> connect(settings_dict=settings_dict, alias=alias)

or
    >>> get_connection(alias, settings_dict=setting_dict)

or
    >>> Connection(settings_dict, alias: Optional[str] = None)

The `alias` string must be unique if more Salesforce databases are used together.
A default alias `None` is acceptable if only one connection is used.

A connection can be found by the alias later:

    >>> connection = get_connection(alias)

Or optionally the `settings_dict` can be repeated, but ignored if the connection
can be found.

Example with multiple databases:

    If we want copy some data between two Salesforce databases:

    >>> db1 = Connection({'USER': user1, 'PASSWORD': ..., 'HOST': ...}, 'salesforce')
    >>> db2 = Connection({'USER': user2, 'PASSWORD': ..., 'HOST': ...}, 'salesforce_2')


Thread safety

The driver is thread safe at DB API level 2:

    >>> threadsafety == 2  # "Threads may share the module and connections."

It is true if we consider a connection identified by an alias, however every
thread uses its own new database connections. A new connection should be opened
in a new thread again with the same `settings_dict` belonging to the same `alias`,
but a connection can not be opened again with the same alias in the same thread.
In other words, connections with the same `alias` in all threads should preferably
use the same `settings_dict`. It is similar to Django where a connection with the
same settings is opened in


Authentication:

The USER name is worldwidey unique in Salesforce SDFC and the Salesforce
database can be unambiguously identified by it.

Many static or dynamic authentication methods can be specified in `settings_dict`
by 'AUTH' key. (see `salesforce.auth` module) If a static authentication method
is used then the same 'USER' user is used in all threads all the time.

It still can be useful to modify the database data on behalf of a particular
currently logged user of a web application.
In a general case a multitenant application is possible with a dynamic
authentication where every thread can frequently switch the database and the user
associated to the `alias` according to session data of a web request. The
connection can become unauthenticated after the end of his/her request.

Threads are intended to prevent a bottle neck on a normal web server, but not
for long running batches in loop in parallel. (Salesforce could maybe reject
such a big load on REST API because a low priority high performance Bulk API
exists instead.)

--

There is also all low level code of the driver for Salesforce here in the module,
not only the DB API.
"""

import logging
from salesforce.dbapi.driver import (  # noqa pylint:disable=useless-import-alias
    Connection as Connection,
    connect as connect,
    get_connection as get_connection,
)
from salesforce.dbapi.exceptions import (  # noqa pylint:disable=useless-import-alias
    IntegrityError as IntegrityError, DatabaseError as DatabaseError, SalesforceError as SalesforceError,
    OperationalError as OperationalError,
)

log = logging.getLogger(__name__)


# -- API global constants

apilevel = "2.0"  # see https://www.python.org/dev/peps/pep-0249
threadsafety = 2

# This paramstyle uses '%s' parameters.
paramstyle = 'format'



from django.test import TestCase
import django.contrib.auth
from salesforce.testrunner.example.models import Account


class WebTest(TestCase):
    databases = '__all__'

    def test_admin(self):
        """Test that mainly improves code coverage."""

        # Log in as a superuser
        user = django.contrib.auth.models.User.objects.create_user('fredsu', 'fred@example.com', 'passwd')
        user.is_superuser = True
        user.is_staff = True
        user.save()
        self.client.login(username='fredsu', password='passwd')

        account = Account(Name='sf_test account')
        account.save()

        response = self.client.get('/')
        response = self.client.get('/search/')
        response = self.client.post('/search/', {'query': 'test account'})
        self.assertIn(b'sf_test account', response.content)
        response = self.client.post('/admin/example/account/')
        response = self.client.post('/admin/example/contact/')
        response = self.client.post('/admin/example/lead/')
        response = self.client.post('/admin/example/pricebook/')
        response = self.client.post('/admin/')
        self.assertIn('PricebookEntries', response.rendered_content)
        account.delete()
        user.delete()



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Salesforce object query and queryset customizations.  (like django.db.models.query)
"""
from typing import Dict, Generic, Iterable, List, NoReturn, Optional, TYPE_CHECKING, Type, TypeVar
import typing  # pylint:disable=unused-import

from django.conf import settings
from django.db import NotSupportedError, models, DEFAULT_DB_ALIAS
from django.db.models import constants
from django.db.models import query as models_query, Model
from django.db.models.sql import where as sql_where
import django

from salesforce.backend.indep import get_sf_alt_pk
from salesforce.backend import compiler, DJANGO_30_PLUS, DJANGO_40_PLUS, DJANGO_41_PLUS
from salesforce.backend.models_sql_query import SalesforceQuery
from salesforce.backend.operations import BULK_BATCH_SIZE
from salesforce.router import is_sf_database
import salesforce.backend.utils

_T = TypeVar("_T", bound=models.Model, covariant=True)
if TYPE_CHECKING:
    from salesforce.models import SalesforceModel  # pylint:disable=cyclic-import # noqa
    # pylint:enable=cyclic-import

# class SalesforceRawQuerySet(query.RawQuerySet):
#     def __len__(self):
#         if self.query.cursor is None:
#             # force the query
#             self.query.get_columns()
#         return self.query.cursor.rowcount

if DJANGO_40_PLUS and not hasattr(sql_where.WhereNode, 'as_salesforce'):
    setattr(sql_where.WhereNode, 'as_salesforce', compiler.SalesforceWhereNode.as_salesforce)


class SalesforceQuerySet(models_query.QuerySet, Generic[_T]):
    """
    Use a custom SQL compiler to generate SOQL-compliant queries.
    """
    if TYPE_CHECKING:
        query = None  # type: SalesforceQuery[_T]

    def __init__(self,
                 model: Optional[Type[_T]] = None,
                 query: Optional['SalesforceQuery[_T]'] = None,
                 using: Optional[str] = None,
                 hints: Optional[Dict[str, Model]] = None
                 ) -> None:
        if query is None:
            if DJANGO_40_PLUS:
                query = SalesforceQuery(model)
            else:
                query = SalesforceQuery(model, where=compiler.SalesforceWhereNode)
        super().__init__(model=model, query=query, using=using, hints=hints)

    def using(self, alias: Optional[str]) -> 'SalesforceQuerySet[_T]':
        if alias is None:
            if hasattr(self.model, '_salesforce_object'):
                alias = getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce')
            else:
                alias = DEFAULT_DB_ALIAS
        return super().using(alias)

    def query_all(self) -> 'SalesforceQuerySet[_T]':
        """
        Allows querying for also deleted or merged records.
            Lead.objects.query_all().filter(IsDeleted=True,...)
        https://www.salesforce.com/us/developer/docs/api_rest/Content/resources_queryall.htm

        It is ignored on non-salesforce databases.
        """
        return self.sf(query_all=True)

    def simple_select_related(self, *fields: str) -> NoReturn:  # pylint:disable=no-self-use
        raise NotSupportedError("Obsoleted method .simple_select_related(), use .select_related() instead")

    def bulk_create(self, objs: Iterable[_T], batch_size: Optional[int] = None,
                    ignore_conflicts: bool = False,
                    update_conflicts: bool = False,
                    update_fields: Optional[List[str]] = None,
                    unique_fields: Optional[List[str]] = None,
                    ) -> List[_T]:
        assert not update_conflicts and update_fields is None and unique_fields is None
        if getattr(self.model, '_salesforce_object', '') == 'extended' and not is_sf_database(self.db):
            objs = list(objs)
            for x in objs:
                if x.pk is None:
                    x.pk = get_sf_alt_pk()
        return super().bulk_create(objs, batch_size=batch_size, ignore_conflicts=ignore_conflicts)

    def bulk_update(self, objs: Iterable[Model], fields: 'typing.Collection[str]',  # pylint:disable=arguments-differ
                    batch_size: Optional[int] = None, all_or_none: bool = None):
        self.sf(all_or_none=all_or_none)
        if batch_size is not None and batch_size < 0:
            raise ValueError('Batch size must be a positive integer.')
        batch_size = min(batch_size, BULK_BATCH_SIZE) if batch_size else BULK_BATCH_SIZE
        for chunk in salesforce.backend.utils.chunked(objs, batch_size):
            bulk_update_small(chunk, fields, all_or_none=all_or_none)

    def sf(self,
           query_all: Optional[bool] = None,
           all_or_none: Optional[bool] = None,
           edge_updates: Optional[bool] = None,
           minimal_aliases: Optional[bool] = None,
           ) -> 'SalesforceQuerySet[_T]':
        """Set additional parameters for queryset methods with Salesforce.

        see details about these parameters in `salesforce.backend.models_sql_query.SalesforceQuery.sf(...)`

        It is better to put this method near the beginning of the chain of queryset methods.

        Example:
        >>> Contact.objects.sf(all_or_none=True).bulk_create([Contact(last_name='a')])
        """
        if not is_sf_database(self.db):
            return self
        clone = self._chain()
        clone.query = clone.query.sf(
            query_all=query_all,
            all_or_none=all_or_none,
            edge_updates=edge_updates,
            minimal_aliases=minimal_aliases,
        )
        return clone

    # def _chain(self, **kwargs) -> 'SalesforceQuerySet[_T]':
    #     return super()._chain(**kwargs)

    def patch_insert_query(self, query: models.sql.Query) -> None:
        setattr(query, 'sf_params', self.query.sf_params)

    # original Django method '._insert(...)' patched by .patch_insert_query(...)
    if DJANGO_41_PLUS:
        def _insert(self, objs, fields,
                    returning_fields=None, raw=False, using=None, on_conflict=None,
                    update_fields=None, unique_fields=None):
            assert on_conflict is None or on_conflict == constants.OnConflict.IGNORE  # pylint:disable=no-member
            assert update_fields is None and unique_fields is None
            self._for_write = True
            if using is None:
                using = self.db
            query = models.sql.InsertQuery(self.model, on_conflict=on_conflict)
            self.patch_insert_query(query)  # patch
            query.insert_values(fields, objs, raw=raw)
            return query.get_compiler(using=using).execute_sql(returning_fields)
    elif DJANGO_30_PLUS:
        def _insert(self, objs, fields,  # type: ignore[misc] # pylint:disable=arguments-differ
                    returning_fields=None, raw=False, using=None, ignore_conflicts=False):
            self._for_write = True
            if using is None:
                using = self.db
            query = models.sql.InsertQuery(self.model, ignore_conflicts=ignore_conflicts)
            self.patch_insert_query(query)  # patch
            query.insert_values(fields, objs, raw=raw)
            return query.get_compiler(using=using).execute_sql(returning_fields)
    else:
        def _insert(self, objs, fields,  # type: ignore[misc] # pylint:disable=arguments-differ
                    return_id=False, raw=False, using=None, ignore_conflicts=False):
            """
            Insert a new record for the given model. This provides an interface to
            the InsertQuery class and is how Model.save() is implemented.
            """
            self._for_write = True
            if using is None:
                using = self.db
            query = models.sql.InsertQuery(self.model, ignore_conflicts=ignore_conflicts)
            self.patch_insert_query(query)  # patch
            query.insert_values(fields, objs, raw=raw)
            return query.get_compiler(using=using).execute_sql(return_id)
    _insert.alters_data = True  # type: ignore[attr-defined]  # noqa
    _insert.queryset_only = False  # type: ignore[attr-defined]  # noqa


def bulk_update_small(objs: 'typing.Collection[models.Model]', fields: Iterable[str], all_or_none: bool = None
                      ) -> None:
    # simple implementation without "batch_size" parameter, but with "all_or_none"
    # and objects from mixed models can be updated by one request in the same transaction
    assert len(objs) <= BULK_BATCH_SIZE
    records = []
    dbs = set()
    for item in objs:
        query = django.db.models.sql.subqueries.UpdateQuery(item._meta.model)  # fake query
        query.add_update_values({field: getattr(item, field) for field in fields})
        values = salesforce.backend.utils.extract_update_values(query)
        values['id'] = item.pk
        values['type_'] = item._meta.db_table
        records.append(values)
        dbs.add(item._state.db)  # pylint:disable=protected-access
    db = dbs.pop()
    if dbs or not is_sf_database(db):
        raise ValueError("All updated objects must be from the same Salesforce database.")
    connection = django.db.connections[db].connection
    connection.sobject_collections_request('PATCH', records, all_or_none=all_or_none)



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

from django.urls import path
from . import views

urlpatterns = [
    path(r'', views.list_accounts, name='list_accounts'),
    path(r'search/', views.search_accounts, name='search_accounts'),
]



# django-salesforce

"""
A set of tools to deal with Salesforce actions that
cannot or can hardly be implemented using the generic
relational database abstraction.

The Salesforce REST API is missing a few endpoints that
are available in the SOAP API. We are using `beatbox` as
a workaround for those specific actions (such as Lead-Contact
conversion).
"""
from typing import Any, Dict, Optional
from django.db import connections

import salesforce
from salesforce.dbapi.driver import beatbox, DatabaseError, InterfaceError


def get_soap_client(db_alias: str, client_class: 'beatbox.PythonClient' = None) -> 'beatbox.PythonClient':
    """
    Create the SOAP client for the current user logged in the db_alias

    The default created client is "beatbox.PythonClient", but an
    alternative client is possible. (i.e. other subtype of beatbox.XMLClient)
    """
    if not beatbox:
        raise InterfaceError("To use SOAP API, you'll need to install the Beatbox package.")
    if client_class is None:
        if hasattr(beatbox, 'PythonClient'):
            client_class = beatbox.PythonClient  # https://github.com/davisagli/beatbox
        else:
            client_class = beatbox.Client        # https://github.com/superfell/Beatbox # pylint:disable=no-member
    soap_client = client_class()

    # authenticate if known that not authenticated
    connection = connections[db_alias]
    # verify the authenticated connection by a request, because Beatbox can not refresh the token
    cursor = connection.cursor()
    cursor.cursor.urls_request()

    auth_info = connection.sf_session.auth

    access_token = auth_info.get_auth()['access_token']
    assert access_token[15] == '!'
    org_id = access_token[:15]
    url = '/services/Soap/u/{version}/{org_id}'.format(version=salesforce.API_VERSION,
                                                       org_id=org_id)
    soap_client.useSession(access_token, auth_info.instance_url + url)
    return soap_client


def convert_lead(lead: Any, converted_status: Optional[str] = None, **kwargs: Any) -> Dict[str, str]:
    """
    Convert `lead` using the `convertLead()` endpoint exposed
    by the SOAP API.

    Parameters:
    `lead` -- a Lead object that has not been converted yet.
    `converted_status` -- valid LeadStatus value for a converted lead.
        Not necessary if only one converted status is configured for Leads.

    kwargs: additional optional parameters according docs
    https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_convertlead.htm
    e.g. `accountId`: if the Lead should be merged with an existing Account.
         `doNotCreateOpportunity`: It specifies whether to create an Opportunity during a Lead conversion

    Return value:
        {'accountId':.., 'contactId':.., 'leadId':.., 'opportunityId':.., 'success':..}

    -- BEWARE --
    The current implementation won't work in case your `Contact`,
    `Account` or `Opportunity` objects have some custom **and**
    required fields. This arises from the fact that `convertLead()`
    is only meant to deal with standard Salesforce fields, so it does
    not really care about populating custom fields at insert time.

    One workaround is to map a custom required field in
    your `Lead` object to every custom required field in the target
    objects (i.e., `Contact`, `Opportunity` or `Account`). Follow the
    instructions at

    https://help.salesforce.com/apex/HTViewHelpDoc?id=customize_mapleads.htm

    for more details.
    """
    # pylint:disable=protected-access
    if not beatbox:
        raise InterfaceError("To use convert_lead, you'll need to install the Beatbox library.")

    accepted_kw = set(('accountId', 'contactId', 'doNotCreateOpportunity',
                       'opportunityName', 'overwriteLeadSource', 'ownerId',
                       'sendNotificationEmail'))
    assert all(x in accepted_kw for x in kwargs)

    db_alias = lead._state.db
    if converted_status is None:
        converted_status = connections[db_alias].introspection.converted_lead_status
    soap_client = get_soap_client(db_alias)

    # convert
    kwargs['leadId'] = lead.pk
    kwargs['convertedStatus'] = converted_status
    response = soap_client.convertLead(kwargs)

    ret = dict((x._name[1], str(x)) for x in response)

    if "errors" in str(ret):
        raise DatabaseError("The Lead conversion failed: {0}, leadId={1}"
                            .format(ret['errors'], ret['leadId']))
    return ret



# minimal dynamic metadata are specified here. Static metadata are in "pyproject.toml".
import os
import setuptools

sf_development = os.path.isfile('.djsf_development')
setuptools.setup(
    install_requires=[
        'django>=2.2' + (',<6.0' if not sf_development else ''),
        'pytz>=2012c',
        'requests>=2.32.0',
    ],
)



"""
Dummy module to be used if models created by `inspectdb` should be a template for a "dynamic model".

A dynamic model is an easy way to start with a simplified customized module and
to can add selected new fields from a complete current model created by
inspectdb, that we call "models template". It is even possible to frequently
refresh the template by inspectdb without confusion or rewriting the already
written customizations.

Example:
    Create a template model by
        python manage.py inspectdb --database=salesforce >your_app/models_template.py

    Replace the line "from salesforce import models"
                  by "from salesforce import models_template as models"

    That means that your `models_template` module is not considered a real
    database models module. It is an ugly file with very long lines.

Dynamic Models created by selection from


"""
from salesforce.models import (  # NOQA pylint:disable=unused-import
        NOT_UPDATEABLE as NOT_UPDATEABLE, NOT_CREATEABLE as NOT_CREATEABLE,
        READ_ONLY as READ_ONLY, DO_NOTHING as DO_NOTHING,
)
from salesforce.models import *  # NOQA pylint:disable=unused-wildcard-import,wildcard-import
from salesforce.backend.indep import LazyField
import salesforce

Model = salesforce.models.ModelTemplate  # type: ignore[assignment,misc]  # noqa

CharField = LazyField(salesforce.models.CharField)
EmailField = LazyField(salesforce.models.EmailField)
URLField = LazyField(salesforce.models.URLField)
TextField = LazyField(salesforce.models.TextField)
IntegerField = LazyField(salesforce.models.IntegerField)
BigIntegerField = LazyField(salesforce.models.BigIntegerField)
SmallIntegerField = LazyField(salesforce.models.SmallIntegerField)
DecimalField = LazyField(salesforce.models.DecimalField)
FloatField = LazyField(salesforce.models.FloatField)
BooleanField = LazyField(salesforce.models.BooleanField)
DateTimeField = LazyField(salesforce.models.DateTimeField)
DateField = LazyField(salesforce.models.DateField)
TimeField = LazyField(salesforce.models.TimeField)
ForeignKey = LazyField(salesforce.models.ForeignKey)
OneToOneField = LazyField(salesforce.models.OneToOneField)
XJSONField = LazyField(salesforce.models.XJSONField)
AutoField = LazyField(salesforce.models.AutoField)  # not important



"""
Customized Query, RawQuery  (like django.db.models.sql.query)
"""
import copy
from typing import Any, cast, Generic, Optional, Sequence, Tuple, Type, TypeVar
from django.conf import settings
from django.db.models import Count, Model
from django.db.models.sql import Query, RawQuery, constants
import django

from salesforce.backend import DJANGO_40_PLUS, DJANGO_42_PLUS
from salesforce.backend.compiler import SfParams, SQLCompiler
from salesforce.dbapi.driver import arg_to_soql

_T = TypeVar("_T", bound=Model, covariant=True)


class SalesforceRawQuery(RawQuery):
    pass

#     def clone(self, using):
#         return SalesforceRawQuery(self.sql, using, params=self.params)
#
#     def get_columns(self):
#         if self.cursor is None:
#             self._execute_query()
#         converter = connections[self.using].introspection.table_name_converter
#         if self.cursor.rowcount > 0:
#             return [converter(col) for col in self.cursor.first_row.keys() if col != 'attributes']
#         # TODO hy: A more general fix is desirable with rewriting more code.
#         return ['Id']
#
#     def _execute_query(self):
#         self.cursor = connections[self.using].cursor()
#         self.cursor.prepare_query(self)
#         self.cursor.execute(self.sql, self.params)
#
#     def __repr__(self):
#         return "<SalesforceRawQuery: %s; %r>" % (self.sql, tuple(self.params))
#
#     def __iter__(self):
#         for row in super().__iter__():
#             yield [row[k] for k in self.get_columns()]


class SalesforceQuery(Query, Generic[_T]):
    """
    Override aggregates.
    """
    def __init__(self, model: Optional[Type[_T]], *args, **kwargs) -> None:
        super().__init__(model, *args, **kwargs)
        self.max_depth = 1
        self.sf_params = SfParams()  # paramaters for Salesforce query instead of transaction control

    def __str__(self) -> str:
        """Return the query as merged SOQL for Salesforce"""
        sql, params = self.sql_with_params()
        return sql % tuple(arg_to_soql(x) for x in params)

    def sql_with_params(self) -> Tuple[str, Sequence[Any]]:
        """
        Return the query as an SOQL string and the parameters for Salesforce.

        It is a shortcut for debugging, unused by backends.
        (It ignores "using(...)". because the exact alias is known only in
        queryset, but not in a query.)
        """
        sf_alias = getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce')
        return self.get_compiler(sf_alias).as_sql()

    def clone(self) -> 'SalesforceQuery[_T]':
        return cast(SalesforceQuery, Query.clone(self))

    def get_compiler(self, using: Optional[str] = None, connection: Any = None,  # pylint:disable=arguments-differ
                     elide_empty: bool = True) -> django.db.models.sql.compiler.SQLCompiler:
        if not DJANGO_40_PLUS:
            compiler = super().get_compiler(using, connection=connection)
        else:
            compiler = super().get_compiler(using, connection=connection,  # pylint:disable=unexpected-keyword-arg
                                            elide_empty=elide_empty)  # type: ignore[call-arg]
        if isinstance(compiler, SQLCompiler):
            return compiler.set_sf_params(self.sf_params)
        return compiler

    def sf(self,
           query_all: Optional[bool] = None,
           all_or_none: Optional[bool] = None,
           edge_updates: Optional[bool] = None,
           minimal_aliases: Optional[bool] = None
           ) -> 'SalesforceQuery[_T]':
        """
        Set additional parameters for a queryset

            `query_all`: True: Also deleted objects from the trash bin can be selected by this.
                These can be selected exclusively after that by `.filter(is_deleted=True)`

            `all_or_none`: This affects `bulk_create` and `bulk_update` methods. These methods
                         work by blocks of 200 items and it is a biggest possible transacion.
                True: All records in blocks are updated or no record is updated in a block with
                    the first error and no following block is tried. (small rolback and stop)
                False: All possible records in all blocks are updated.
                Default: No following block after an error is tried, but also no rollback is done
                    in a block with some errors. (This neutral behavior can not be set back after
                    True or False.)

            `edge_updates`: methods update() and delete() on querysets with related tables
                could be unsafe if the queryset is not checked. It is safe to rewrite it to two
                nested querysets or if it is correct then if can be allowed by `edge_updates`.
                The default value is False.

            `minimal_aliases`: Fields are compiled to a simple "field_name" if pssible without a dot,
                not to a "table_alias.field_name".
        """
        clone = self.clone()
        clone.sf_params = copy.copy(self.sf_params)
        if query_all is not None:
            clone.sf_params.query_all = query_all
        if all_or_none is not None:
            clone.sf_params.all_or_none = all_or_none
        if edge_updates is not None:
            clone.sf_params.edge_updates = edge_updates
        if minimal_aliases is not None:
            clone.sf_params.minimal_aliases = minimal_aliases
        return clone

    def has_results(self, using: Optional[str]) -> bool:
        q = self.clone()
        compiler = q.get_compiler(using=using)
        return bool(compiler.execute_sql(constants.SINGLE))

    def get_count(self, using: str) -> int:
        """
        Performs a COUNT() query using the current filter constraints.
        """
        # customized because "Count('*')" is not possible with Salesforce and also not "__" in an alias
        obj = self.clone()
        if DJANGO_42_PLUS:
            return obj.get_aggregation(using, {"x_sf_count": Count("*")})["x_sf_count"]
        obj.add_annotation(Count('pk'), alias='x_sf_count', is_summary=True)  # pylint: disable=no-member
        number = obj.get_aggregation(using, ['x_sf_count'])['x_sf_count']  # type: Optional[int] # pylint: disable=no-member # noqa
        if number is None:
            number = 0
        return number



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Salesforce database backend for Django.  (like 'django.db.backends.*.base')
"""

from typing import Any, Dict, Optional, TYPE_CHECKING

from django.conf import settings
from django.db.backends.base.base import BaseDatabaseWrapper

from salesforce.backend import enterprise
from salesforce.backend.client import DatabaseClient
from salesforce.backend.creation import DatabaseCreation
from salesforce.backend.features import DatabaseFeatures
from salesforce.backend.validation import DatabaseValidation
from salesforce.backend.operations import DatabaseOperations
from salesforce.backend.introspection import DatabaseIntrospection
from salesforce.backend.schema import DatabaseSchemaEditor
# from django.db.backends.signals import connection_created
from salesforce.backend.utils import CursorWrapper, async_unsafe
from salesforce.dbapi import driver as Database
from salesforce.dbapi.driver import IntegrityError, DatabaseError, SalesforceError  # NOQA pylint:disable=unused-import

if TYPE_CHECKING:
    from django.db.backends.base.base import ProtoCursor  # pylint:disable=ungrouped-imports,no-name-in-module


__all__ = ('DatabaseWrapper', 'DatabaseError', 'SalesforceError',)


class DatabaseWrapper(BaseDatabaseWrapper):
    """
    Core class that provides all DB support.
    """
    # pylint:disable=abstract-method,too-many-instance-attributes
    #     undefined abstract methods: _start_transaction_under_autocommit, is_usable

    vendor = 'salesforce'
    display_name = 'Salesforce'

    # All string operators are case insensitive in SOQL. (that can not be fixed)
    # The relevant part here is the operator, e.g. a simple "LIKE".
    # An optional wilcard ('%') is managed by a universal Django code for an operand.
    operators = {
        'exact': '= %s',
        'iexact': 'LIKE %s',
        'contains': 'LIKE %s',
        'icontains': 'LIKE %s',
        # 'regex': 'REGEXP %s',  # unsupported
        # 'iregex': 'REGEXP %s',
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': 'LIKE %s',
        'endswith': 'LIKE %s',
        'istartswith': 'LIKE %s',
        'iendswith': 'LIKE %s',
    }

    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor  # type: ignore[assignment] # noqa # this is normal in Django

    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    validation_class = DatabaseValidation

    def __init__(self, settings_dict, alias=None):
        if alias is None:
            alias = getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce')
        super().__init__(settings_dict, alias)

        self._is_sandbox = None  # type: Optional[bool]

    @property
    def sf_session(self) -> Database.SfSession:
        if self.connection is None:
            self.connect()
            assert self.connection
        return self.connection.sf_session

    def get_connection_params(self) -> Dict[str, Any]:
        settings_dict = self.settings_dict
        params = settings_dict.copy()
        params.update(settings_dict['OPTIONS'])
        return params

    @async_unsafe
    def get_new_connection(self, conn_params: Dict[str, Any]) -> Database.RawConnection:
        enterprise.check_license_in_latest_django()
        # simulated only a connection interface without connecting really
        return Database.connect(settings_dict=conn_params, alias=self.alias)

    def init_connection_state(self):
        pass  # nothing to init

    def _set_autocommit(self, autocommit):
        # SF REST API uses autocommit, but until rollback it is not a
        # serious problem to ignore autocommit off
        pass

    @async_unsafe
    def cursor(self) -> Any:
        """
        Return a fake cursor for accessing the Salesforce API with SOQL.
        """
        return CursorWrapper(self.create_cursor(), self)

    def create_cursor(self, name: Optional[str] = None) -> 'ProtoCursor':
        row_type = {'dict': dict, 'list': list, None: None}[name]
        self.sf_session  # to connect
        return self.connection.cursor(row_type=row_type)

    @property
    def is_sandbox(self) -> bool:
        if self._is_sandbox is None:
            cur = self.cursor()
            cur.execute("SELECT IsSandbox FROM Organization")
            self._is_sandbox = cur.fetchone()[0]
        return self._is_sandbox

    def close(self) -> None:
        if self.connection:
            self.connection.close()



"""This is a part of an automatically exported database schema ('inspectdb')

Shortened due to readability:
removed very long choices, removed many long fields, reformated long
lines, but still an example of a long model exported from Salesforce.
"""
from salesforce import models_template as models


class User(models.Model):
    Username = models.CharField(max_length=80)
    Email = models.CharField(max_length=100)
    LastName = models.CharField(max_length=80)
    FirstName = models.CharField(max_length=40)


class Organization(models.Model):
    name = models.CharField(max_length=80, sf_read_only=models.NOT_CREATEABLE)
    division = models.CharField(max_length=80, sf_read_only=models.NOT_CREATEABLE, blank=True)
    street = models.TextField(sf_read_only=models.NOT_CREATEABLE, blank=True)
    city = models.CharField(max_length=40, sf_read_only=models.NOT_CREATEABLE, blank=True)
    country = models.CharField(max_length=80, sf_read_only=models.READ_ONLY, blank=True)
    address = models.TextField(sf_read_only=models.READ_ONLY, blank=True)  # This field type is a guess.
    phone = models.CharField(max_length=40, sf_read_only=models.NOT_CREATEABLE, blank=True)
    instance_name = models.CharField(max_length=5, sf_read_only=models.READ_ONLY, blank=True)
    is_sandbox = models.BooleanField(sf_read_only=models.READ_ONLY)
    created_date = models.DateTimeField(sf_read_only=models.READ_ONLY)
    created_by = models.ForeignKey(User, related_name='organization_createdby_set',
                                   sf_read_only=models.READ_ONLY, on_delete=models.DO_NOTHING)
    last_modified_date = models.DateTimeField(sf_read_only=models.READ_ONLY)
    last_modified_by = models.ForeignKey('User', related_name='organization_lastmodifiedby_set',
                                         sf_read_only=models.READ_ONLY, on_delete=models.DO_NOTHING)

    class Meta(models.Model.Meta):
        db_table = 'Organization'
        verbose_name = 'Organization'
        verbose_name_plural = 'Organizations'
        # keyPrefix = '00D'



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

from typing import Optional
import types

import django
from django.conf import settings

from salesforce import models
from salesforce.backend import DJANGO_50_PLUS
if not getattr(settings, 'SF_EXAMPLE_EXTENDED_MODELS', False):
    from salesforce.models import SalesforceModel as SalesforceModelParent
else:
    from salesforce.models_extend import SalesforceModel as SalesforceModelParent


SALUTATIONS = [
    'Mr.', 'Ms.', 'Mrs.', 'Dr.', 'Prof.'
]

INDUSTRIES = [
    'Agriculture', 'Apparel', 'Banking', 'Biotechnology', 'Chemicals',
    'Communications', 'Construction', 'Consulting', 'Education',
    'Electronics', 'Energy', 'Engineering', 'Entertainment', 'Environmental',
    'Finance', 'Food & Beverage', 'Government', 'Healthcare', 'Hospitality',
    'Insurance', 'Machinery', 'Manufacturing', 'Media', 'Not For Profit',
    'Other', 'Recreation', 'Retail', 'Shipping', 'Technology',
    'Telecommunications', 'Transportation', 'Utilities'
]

# pylint:disable=invalid-str-returned  # return type of __str__ method


# This class customizes `managed = True` for tests and does not disturbe SF
class SalesforceModel(SalesforceModelParent):
    class Meta:
        abstract = True
        managed = True


class User(SalesforceModel):
    Username = models.CharField(max_length=80)
    Email = models.CharField(max_length=100)
    LastName = models.CharField(max_length=80)
    FirstName = models.CharField(max_length=40)
    IsActive = models.BooleanField(default=False)
    UserType = models.CharField(max_length=80, default='Standard')


class AbstractAccount(SalesforceModel):
    """
    Default Salesforce Account model.
    """
    TYPES = [
        'Analyst', 'Competitor', 'Customer', 'Integrator', 'Investor',
        'Partner', 'Press', 'Prospect', 'Reseller', 'Other'
    ]

    Owner = models.ForeignKey(User, on_delete=models.DO_NOTHING,
                              default=models.DefaultedOnCreate(User),
                              db_column='OwnerId')
    Type = models.CharField(max_length=100, choices=[(x, x) for x in TYPES],
                            null=True)
    BillingStreet = models.CharField(max_length=255)
    BillingCity = models.CharField(max_length=40)
    BillingState = models.CharField(max_length=20)
    BillingPostalCode = models.CharField(max_length=20)
    BillingCountry = models.CharField(max_length=40)
    ShippingStreet = models.CharField(max_length=255)
    ShippingCity = models.CharField(max_length=40)
    ShippingState = models.CharField(max_length=20)
    ShippingPostalCode = models.CharField(max_length=20)
    ShippingCountry = models.CharField(max_length=40)
    Phone = models.CharField(max_length=255)
    Website = models.CharField(max_length=255)
    Industry = models.CharField(max_length=100,
                                choices=[(x, x) for x in INDUSTRIES])
    Description = models.TextField()
    # Added read only option, otherwise the object can not be never saved
    # If the model is used also with non SF databases then there should be set
    # allow_now=True or null=True
    LastModifiedDate = models.DateTimeField(db_column='LastModifiedDate',
                                            sf_read_only=models.READ_ONLY,
                                            auto_now=True)

    class Meta(SalesforceModel.Meta):
        abstract = True

    def __str__(self):
        return self.Name  # pylint: disable=no-member


class CoreAccount(AbstractAccount):
    Name = models.CharField(max_length=255)

    class Meta(AbstractAccount.Meta):
        abstract = True


class PersonAccount(AbstractAccount):
    # Non standard fields that require activating "Person Account"
    # (irreversible changes in Salesforce)
    LastName = models.CharField(max_length=80)
    FirstName = models.CharField(max_length=40)
    Name = models.CharField(max_length=255, sf_read_only=models.READ_ONLY)
    Salutation = models.CharField(max_length=100,
                                  choices=[(x, x) for x in SALUTATIONS])
    IsPersonAccount = models.BooleanField(default=False, sf_read_only=models.READ_ONLY)
    PersonEmail = models.CharField(max_length=100)

    class Meta(AbstractAccount.Meta):
        abstract = True


if getattr(settings, 'SF_EXAMPLE_PERSON_ACCOUNT_ACTIVATED', False):
    class Account(PersonAccount):  # pylint:disable=model-no-explicit-unicode
        pass
else:
    class Account(CoreAccount):  # type: ignore[no-redef]  # noqa # pylint:disable=model-no-explicit-unicode
        pass


class Contact(SalesforceModel):
    # Example that db_column is not necessary for most of fields even with
    # lower case names and for ForeignKey
    account = models.ForeignKey(Account, on_delete=models.DO_NOTHING,
                                blank=True, null=True)  # db_column: 'AccountId'
    last_name = models.CharField(max_length=80)
    first_name = models.CharField(max_length=40, blank=True, null=True)
    name = models.CharField(max_length=121, sf_read_only=models.READ_ONLY,
                            verbose_name='Full Name')
    email = models.EmailField(blank=True, null=True)
    email_bounced_date = models.DateTimeField(blank=True, null=True)
    # The `default=` with lambda function is easy readable, but can be
    # problematic with migrations in the future because it is not serializable.
    # It can be replaced by normal function.
    if not getattr(settings, 'SF_EXAMPLE_SIMPLE_DEFAULTS', False):
        kw_owner = {('db_default' if DJANGO_50_PLUS else 'default'): models.DEFAULTED_ON_CREATE}
        owner = models.ForeignKey(User, on_delete=models.DO_NOTHING,
                                  related_name='contact_owner_set', blank=True,
                                  **kw_owner
                                  )
    if getattr(settings, 'SF_EXAMPLE_CUSTOM_INSTALLED', False):
        vs = models.DecimalField(custom=True, unique=True, max_digits=10, decimal_places=0, blank=True, null=True)

    def __str__(self):
        return self.name


class Lead(SalesforceModel):
    """
    Default Salesforce Lead model.
    """
    SOURCES = [
        'Advertisement', 'Employee Referral', 'External Referral',
        'Partner', 'Public Relations',
        'Seminar - Internal', 'Seminar - Partner', 'Trade Show', 'Web',
        'Word of mouth', 'Other',
    ]

    STATUSES = [
        'Contacted', 'Open', 'Qualified', 'Unqualified',
    ]

    RATINGS = [
        'Hot', 'Warm', 'Cold',
    ]

    LastName = models.CharField(max_length=80)
    FirstName = models.CharField(max_length=40, blank=True, null=True)
    Salutation = models.CharField(max_length=100, choices=[(x, x) for x in SALUTATIONS])
    Salutation = models.CharField(max_length=100,
                                  choices=[(x, x) for x in SALUTATIONS])
    Name = models.CharField(max_length=121, sf_read_only=models.READ_ONLY)
    Title = models.CharField(max_length=128)
    Company = models.CharField(max_length=255)
    Street = models.CharField(max_length=255)
    City = models.CharField(max_length=40)
    State = models.CharField(max_length=20)
    PostalCode = models.CharField(max_length=20)
    Country = models.CharField(max_length=40)
    Phone = models.CharField(max_length=255)
    Email = models.CharField(max_length=100)
    LeadSource = models.CharField(max_length=100,
                                  choices=[(x, x) for x in SOURCES])
    Status = models.CharField(max_length=100, choices=[(x, x) for x in STATUSES])
    Industry = models.CharField(max_length=100,
                                choices=[(x, x) for x in INDUSTRIES])
    # Added an example of special DateTime field in Salesforce that can
    # not be inserted, but can be updated
    # TODO write test for it
    EmailBouncedDate = models.DateTimeField(blank=True, null=True,
                                            sf_read_only=models.NOT_CREATEABLE)
    # Deleted object can be found only in querysets with "query_all" SF method.
    IsDeleted = models.BooleanField(default=False, sf_read_only=models.READ_ONLY)
    if not getattr(settings, 'SF_EXAMPLE_SIMPLE_DEFAULTS', False):
        kw_owner = {('db_default' if DJANGO_50_PLUS else 'default'): models.DEFAULTED_ON_CREATE}
        owner = models.ForeignKey(User, on_delete=models.DO_NOTHING,
                                  related_name='lead_owner_set', **kw_owner)
    last_modified_by = models.ForeignKey(User, on_delete=models.DO_NOTHING, null=True,
                                         sf_read_only=models.READ_ONLY,
                                         related_name='lead_lastmodifiedby_set')
    is_converted = models.BooleanField(verbose_name='Converted',
                                       sf_read_only=models.NOT_UPDATEABLE,
                                       default=models.DEFAULTED_ON_CREATE)

    def __str__(self):
        return self.Name


class Product(SalesforceModel):
    Name = models.CharField(max_length=255)

    class Meta(SalesforceModel.Meta):
        db_table = 'Product2'

    def __str__(self):
        return self.Name


class Pricebook(SalesforceModel):
    Name = models.CharField(max_length=255)

    class Meta(SalesforceModel.Meta):
        db_table = 'Pricebook2'

    def __str__(self):
        return self.Name


class PricebookEntry(SalesforceModel):
    Name = models.CharField(max_length=255, db_column='Name', sf_read_only=models.READ_ONLY)
    Pricebook2 = models.ForeignKey('Pricebook', on_delete=models.DO_NOTHING)
    Product2 = models.ForeignKey('Product', on_delete=models.DO_NOTHING)
    UseStandardPrice = models.BooleanField(default=False)
    UnitPrice = models.DecimalField(decimal_places=2, max_digits=18)

    class Meta(SalesforceModel.Meta):
        db_table = 'PricebookEntry'
        verbose_name_plural = "PricebookEntries"

    def __str__(self):
        return self.Name


class ChargentOrder(SalesforceModel):
    # the class is used only by unit tests, it is not necessary to be installed
    class Meta(SalesforceModel.Meta):
        db_table = 'ChargentOrders__ChargentOrder__c'
        custom = True
        managed = False  # can not be managed if it eventually could not exist

    Name = models.CharField(max_length=255, db_column='Name')
    Balance_Due = models.CharField(max_length=255, db_column='ChargentOrders__Balance_Due__c')


class CronTrigger(SalesforceModel):
    # A special DateTime field with milisecond resolution (read only)
    PreviousFireTime = models.DateTimeField(verbose_name='Previous Run Time', blank=True, null=True)
    # ...


class BusinessHours(SalesforceModel):
    Name = models.CharField(max_length=80)
    # The default record is automatically created by Salesforce.
    IsDefault = models.BooleanField(default=False, verbose_name='Default Business Hours')
    # ... much more fields, but we use only this one TimeFiled for test
    MondayStartTime = models.TimeField()

    class Meta:
        verbose_name_plural = "BusinessHours"


class SalesforceParentModel(SalesforceModel):
    """
    Example of standard fields present in all custom models.
    """
    # This is not a custom field because is not defined in a custom model.
    # The API name is therefore 'Name'.
    name = models.CharField(max_length=80)
    last_modified_date = models.DateTimeField(sf_read_only=models.READ_ONLY, auto_now=True)
    # This model is not custom because it has not an explicit attribute
    # `custom = True` in Meta and also has not a `db_table` that ends with
    # '__c'.

    class Meta:
        abstract = True


class Note(SalesforceModel):
    title = models.CharField(max_length=80)
    body = models.TextField(null=True)
    parent_id = models.CharField(max_length=18)
    parent_type = models.CharField(max_length=50, db_column='Parent.Type', sf_read_only=models.READ_ONLY)


class Opportunity(SalesforceModel):
    name = models.CharField(max_length=255)
    contacts = django.db.models.ManyToManyField(
        Contact, through='example.OpportunityContactRole', related_name='opportunities'
    )
    close_date = models.DateField()
    stage = models.CharField(max_length=255, db_column='StageName')  # e.g. "Prospecting"
    created_date = models.DateTimeField(sf_read_only=models.READ_ONLY, auto_now_add=True)
    amount = models.DecimalField(max_digits=18, decimal_places=2, blank=True, null=True)
    probability = models.DecimalField(
        max_digits=3, decimal_places=0, verbose_name='Probability (%)',
        default=models.DEFAULTED_ON_CREATE, blank=True, null=True)


class OpportunityContactRole(SalesforceModel):
    opportunity = models.ForeignKey(Opportunity, on_delete=models.DO_NOTHING, related_name='contact_roles')
    contact = models.ForeignKey(Contact, on_delete=models.DO_NOTHING, related_name='opportunity_roles')
    role = models.CharField(max_length=40, blank=True, null=True)  # e.g. "Business User"


class OpportunityLineItem(SalesforceModel):
    opportunity = models.ForeignKey(Opportunity, on_delete=models.DO_NOTHING)
    pricebook_entry = models.ForeignKey('PricebookEntry', models.DO_NOTHING, verbose_name='Price Book Entry ID',
                                        sf_read_only=models.NOT_UPDATEABLE)
    product2 = models.ForeignKey('Product', models.DO_NOTHING, verbose_name='Product ID',
                                 sf_read_only=models.NOT_UPDATEABLE)
    name = models.CharField(max_length=376, verbose_name='Opportunity Product Name', sf_read_only=models.READ_ONLY)
    quantity = models.DecimalField(max_digits=12, decimal_places=2)
    total_price = models.DecimalField(max_digits=18, decimal_places=2, default=models.DEFAULTED_ON_CREATE)
    unit_price = models.DecimalField(max_digits=18, decimal_places=2, verbose_name='Sales Price',
                                     default=models.DEFAULTED_ON_CREATE)


try:
    models_template: Optional[types.ModuleType] = None
    from salesforce.testrunner.example import models_template
except ImportError:
    # this is useful for the case that the model is being rewritten by inspectdb
    models_template = None


class Organization(SalesforceModel):
    name = models.CharField(max_length=80, sf_read_only=models.NOT_CREATEABLE)
    division = models.CharField(max_length=80, sf_read_only=models.NOT_CREATEABLE, blank=True)
    organization_type = models.CharField(max_length=40, verbose_name='Edition',
                                         sf_read_only=models.READ_ONLY
                                         )  # e.g 'Developer Edition', Enteprise, Unlimited...
    instance_name = models.CharField(max_length=5, sf_read_only=models.READ_ONLY, blank=True)
    is_sandbox = models.BooleanField(sf_read_only=models.READ_ONLY)
    # Fields created_by, last_modified_by, last_modified_date are dynamic

    class Meta:
        db_table = 'Organization'
        # Copy all fields that match the patters for Force.com field name
        # from the class that use the same db_table "Organization" in the
        # module models_template
        if models_template:
            dynamic_field_patterns = models_template, ['created_by', 'last.*_by']


class Test(SalesforceParentModel):
    """
    Simple custom model with one custom and more standard fields.

    Salesforce object for this model can be created:
    A) automatically from the branch hynekcer/tooling-api-and-metadata
       by commands:
        $ python manage.py shell
            >> from salesforce.backend import tooling
            >> tooling.install_metadata_service()
            >> tooling.create_demo_test_object()
    or
    B) manually can create the same object with `API Name`: `django_Test__c`
        `Data Type` of the Record Name: `Text`

       Create three fields:
       Type            | API Name | Label
       ----------------+----------+----------
       Text            | TestText | Test Text
       Checkbox        | TestBool | Test Bool
       Lookup(Contact) | Contact  | Contact

       Set it accessible by you. (`Set Field-Leved Security`)
    """
    # This is a custom field because it is defined in the custom model.
    # The API name is therefore 'TestField__c'
    test_text = models.CharField(max_length=40, db_column='TestText__c')
    test_bool = models.BooleanField(default=False, db_column='TestBool__c')
    contact = models.ForeignKey(Contact, null=True, on_delete=models.DO_NOTHING, custom=True)

    class Meta:
        custom = True
        db_table = 'django_Test__c'


class TestDetail(SalesforceModel):
    parent = models.ForeignKey(Test, models.DO_NOTHING, db_column='Parent__c', sf_read_only=models.NOT_UPDATEABLE)
    contact = models.ForeignKey('Contact', models.DO_NOTHING, db_column='Contact__c', blank=True, null=True)

    class Meta(SalesforceModel.Meta):
        db_table = 'django_Test_detail__c'


# example of relationship from builtin object to custom object

class Attachment(SalesforceModel):
    # A standard SFDC object that can have a relationship to any custom object
    name = models.CharField(max_length=80)
    parent = models.ForeignKey(Test, sf_read_only=models.NOT_UPDATEABLE, on_delete=models.DO_NOTHING)
    # The "body" of Attachment can't be queried for more rows togehter.
    body = models.TextField()


class Task(SalesforceModel):
    # Reference to tables [Contact, Lead]
    who = models.ForeignKey(Lead, on_delete=models.DO_NOTHING, blank=True, null=True)
    # Refer
    what = models.ForeignKey(Account, related_name='task_what_set', on_delete=models.DO_NOTHING, blank=True, null=True)


# OneToOneField

class ApexEmailNotification(SalesforceModel):
    """Stores who should be notified when unhandled Apex exceptions occur.

    The target can be more Salesforce users or an external email addresses.
    Available in API version 35.0 and later.
    """
    # A semicolon-delimited list of email addresses to notify when unhandled Apex exceptions occur.
    user = models.OneToOneField('User', related_name='apex_email_notification',
                                on_delete=models.DO_NOTHING, blank=True, null=True)
    # Users of your org to notify when unhandled Apex exceptions occur.
    email = models.CharField(unique=True, max_length=255, verbose_name='email', blank=True)


class Campaign(SalesforceModel):
    name = models.CharField(max_length=80)
    number_sent = models.DecimalField(max_digits=18, decimal_places=0, verbose_name='Num Sent', blank=True, null=True)
    parent = models.ForeignKey('Campaign', on_delete=models.DO_NOTHING, blank=True, null=True)
    start_date = models.DateField(blank=True, null=True)
    created_date = models.DateTimeField(sf_read_only=models.READ_ONLY, auto_now_add=True)


class CampaignMember(SalesforceModel):
    campaign = models.ForeignKey(Campaign, on_delete=models.DO_NOTHING)
    contact = models.ForeignKey(Contact, on_delete=models.DO_NOTHING)
    status = models.CharField(max_length=40)


# this model will be removed to test removing in migrations
class DeletedObject(SalesforceModel):
    pass


# models for work with file attachments of objects

class ContentDocument(models.Model):
    owner_id = models.CharField(max_length=18, sf_read_only=models.NOT_CREATEABLE)
    title = models.CharField(max_length=255, sf_read_only=models.NOT_CREATEABLE)
    latest_published_version = models.ForeignKey('ContentVersion', models.DO_NOTHING, sf_read_only=models.READ_ONLY,
                                                 blank=True, null=True)
    description = models.TextField(sf_read_only=models.NOT_CREATEABLE, blank=True, null=True)
    content_size = models.IntegerField(sf_read_only=models.READ_ONLY, blank=True, null=True)
    file_type = models.CharField(max_length=20, sf_read_only=models.READ_ONLY, blank=True, null=True)
    file_extension = models.CharField(max_length=40, sf_read_only=models.READ_ONLY, blank=True, null=True)


class ContentDocumentLink(models.Model):
    linked_entity_id = models.CharField(max_length=18, sf_read_only=models.NOT_UPDATEABLE)
    content_document = models.ForeignKey(ContentDocument, models.DO_NOTHING, sf_read_only=models.NOT_UPDATEABLE)
    share_type = models.CharField(max_length=40, blank=True, null=True,
                                  choices=[('V', 'Viewer'), ('C', 'Collaborator'), ('I', 'Inferred')])
    visibility = models.CharField(max_length=40, blank=True, null=True,
                                  choices=[('AllUsers', 'All Users'), ('InternalUsers', 'Standard Users'),
                                           ('SharedUsers', 'Shared Users')])


class ContentVersion(models.Model):
    content_document = models.ForeignKey(ContentDocument, models.DO_NOTHING, sf_read_only=models.NOT_UPDATEABLE)
    content_url = models.URLField(blank=True, null=True)
    version_number = models.CharField(max_length=20, sf_read_only=models.READ_ONLY, blank=True, null=True)
    title = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    path_on_client = models.CharField(max_length=500, sf_read_only=models.NOT_UPDATEABLE, blank=True, null=True)
    owner = models.ForeignKey(User, models.DO_NOTHING, db_default=models.DEFAULTED_ON_CREATE, blank=True)
    file_type = models.CharField(max_length=20, sf_read_only=models.READ_ONLY)
    version_data = models.TextField(blank=True, null=True)
    content_size = models.IntegerField(sf_read_only=models.READ_ONLY, blank=True, null=True)
    file_extension = models.CharField(max_length=40, sf_read_only=models.READ_ONLY, blank=True, null=True)
    origin = models.CharField(max_length=40, sf_read_only=models.NOT_UPDATEABLE, default='C',
                              choices=[('C', 'Content'), ('H', 'Chatter')])
    content_location = models.CharField(max_length=40, sf_read_only=models.NOT_UPDATEABLE, default='S',
                                        choices=[('S', 'Salesforce'), ('E', 'External'),
                                                 ('L', 'Social Customer Service')])
    version_data_url = models.CharField(max_length=255, sf_read_only=models.READ_ONLY, blank=True, null=True)
